import tkinter as tk
from pyrustic import widget
from pyrustic import tkmisc
from pyrustic.view import View
from pyrustic.tkmisc import merge_cnfs


# Components
LABEL_HEADER = "label_header"
LABEL_MESSAGE = "label_message"
FRAME_FOOTER = "frame_footer"
BUTTON_CANCEL = "button_cancel"
BUTTON_CONFIRM = "button_confirm"


class Confirm(widget.Toplevel):
    """
    Confirm is a dialog box to ask the user to confirm an action.

    Example:

        import tkinter as tk
        from pyrustic.widget.confirm import Confirm

        def my_handler(result):
            print(result)

        root = tk.Tk()
        confirm = Confirm(root, title="Confirm", header="Confirmation",
                        message="Do you really want to continue ?",
                        handler=my_handler)
        confirm.build()
        root.mainloop()

    """
    def __init__(self,
                 master=None,
                 title=None,
                 header=None,
                 message=None,
                 handler=None,
                 geometry=None,
                 cnfs=None):
        """
        PARAMETERS:

        - master: widget parent. Example: an instance of tk.Frame

        - title: title of dialog box

        - header: the text to show as header

        - message: the text to show as message

        - handler: a callback to be executed immediately after closing the dialog box.
            This callback should accept a boolean positional argument.
            True means Ok, confirmed.

        - geometry: str, as the dialog box is a toplevel (BODY),
         you can edit its geometry. Example: "500x300"

        - options: dictionary of widgets options
            The widgets keys are: BODY, LABEL_HEADER,
             LABEL_MESSAGE, FRAME_FOOTER, BUTTON_CANCEL, BUTTON_CONFIRM.

            Example: Assume that you want to set the LABEL_MESSAGE's background to black
            and the BODY's background to red:
                options = { BODY: {"background": "red"},
                            LABEL_MESSAGE: {"background": "black"} }

        """
        self.__cnfs = merge_cnfs(None, cnfs, components=("body",
                        LABEL_HEADER, LABEL_MESSAGE, FRAME_FOOTER,
                        BUTTON_CANCEL, BUTTON_CONFIRM))
        super().__init__(master=master,
                         class_="Confirm",
                         cnf=self.__cnfs["body"],
                         on_build=self.__on_build,
                         on_display=self.__on_display,
                         on_destroy=self.__on_destroy,
                         toplevel_geometry=self.__toplevel_geometry)
        self.__title = title
        self.__header = header
        self.__message = message
        self.__handler = handler
        self.__geometry = geometry
        self.__components = {}
        self.__ok = False
        # build
        self.__view = self.build()

    # ====================================
    #           PROPERTIES
    # ====================================

    @property
    def header(self):
        return self.__header

    @property
    def message(self):
        return self.__message

    @property
    def handler(self):
        return self.__handler

    @property
    def ok(self):
        """
        Returns True if user confirmed, else get False
        """
        return self.__ok

    @property
    def components(self):
        """
        Get the components (widgets instances) used to build this dialog.

        This property returns a dict. The keys are:
            BODY, LABEL_HEADER,
            LABEL_MESSAGE, FRAME_FOOTER, BUTTON_CANCEL, BUTTON_CONFIRM

        Warning: check the presence of key before usage
        """
        return self.__components

    # ====================================
    #               INTERNAL
    # ====================================
    def __on_build(self):
        self.title(self.__title)
        self.resizable(0, 0)
        #
        #
        if self.__geometry:
            self.geometry(self.__geometry)
        #
        if self.__header:
            label_header = tk.Label(self,
                                    text=self.__header,
                                    anchor="w",
                                    justify=tk.LEFT,
                                    name=LABEL_HEADER,
                                    cnf=self.__cnfs[LABEL_HEADER])
            self.__components[LABEL_HEADER] = label_header
            label_header.pack(fill=tk.X, expand=1, anchor="w", pady=5, padx=5)
        #
        if self.__message:
            label_message = tk.Label(self,
                                     name=LABEL_MESSAGE,
                                     text=self.__message,
                                     anchor="w",
                                     justify=tk.LEFT,
                                     cnf=self.__cnfs[LABEL_MESSAGE])
            self.__components[LABEL_MESSAGE] = label_message
            label_message.pack(fill=tk.BOTH,
                               expand=1, padx=5, pady=(5, 10))

        #
        frame_footer = tk.Frame(self, cnf=self.__cnfs[FRAME_FOOTER])
        self.__components[FRAME_FOOTER] = frame_footer
        frame_footer.pack(anchor="e", pady=(0, 2), padx=2)
        #
        button_confirm = tk.Button(frame_footer,
                                   text="Confirm",
                                   name=BUTTON_CONFIRM,
                                   command=self.__on_click_confirm,
                                   cnf=self.__cnfs[BUTTON_CONFIRM])
        self.__components[BUTTON_CONFIRM] = button_confirm
        button_confirm.pack(side=tk.RIGHT)
        #
        button_cancel = tk.Button(frame_footer,
                                  text="Cancel",
                                  name=BUTTON_CANCEL,
                                  command=self.__on_click_cancel,
                                  cnf=self.__cnfs[BUTTON_CANCEL])
        self.__components[BUTTON_CANCEL] = button_cancel
        button_cancel.pack(side=tk.RIGHT, padx=(0, 2))

    def __on_display(self):
        pass

    def __on_destroy(self):
        if self.__handler:
            self.__handler(self.__ok)

    def __toplevel_geometry(self):
        tkmisc.center_window(self, within=self.master.winfo_toplevel())
        tkmisc.dialog_effect(self)

    def __on_click_cancel(self):
        self.__ok = False
        self.destroy()

    def __on_click_confirm(self):
        self.__ok = True
        self.destroy()


class _ConfirmTest(View):
    def __init__(self, root):
        super().__init__()
        self._root = root
        self._body = None

    def _on_build(self):
        self._body = tk.Frame(self._root)
        btn_launch = tk.Button(self._body, text="Launch",
                               command=self._on_click_launch)
        btn_launch.pack()

    def _on_display(self):
        pass

    def _on_destroy(self):
        pass

    def _on_click_launch(self):
        confirm = Confirm(root, title="Confirm",
                          header="Confirmation",
                          message="Do you really want to continue ?\nPress ok to continue\nOr die !")
        confirm.wait_window()
        print("Confirm:", confirm.ok)


if __name__ == "__main__":
    root = tk.Tk()
    root.geometry("500x300+0+0")
    confirm_test = _ConfirmTest(root)
    confirm_test.build_pack()
    root.mainloop()



import os
import os.path
import sys
import subprocess
import getpass
from pyrustic.jasonix import Jasonix
from pyrustic.manager.hubway import Catapult
from pyrustic.manager.misc.funcs import create_gurl, setup_config


class PublishHandler:
    """
    Description
    -----------
    Use this command to build a distribution package
    that could be published later with Hub.
    This command will block the Pyrustic Manager.
    The distribution package is a Wheel.

    Usage
    -----
    - Description: Build
    - Command: build
    """

    def __init__(self, target, app_pkg):
        self._target = target
        self._app_pkg = app_pkg
        self._version = None
        self._pre_publishing_hook = None
        self._post_publishing_hook = None
        self._process(target, app_pkg)

    def _process(self, target, app_pkg):
        if target is None:
            print("Please link a Target first. Check 'help target'.")
            return
        if not app_pkg:
            print("Please initialize this project first. Check 'help init'.")
            return
        # get version
        self._version = self._get_version()
        # get confirmation
        message = "You are going to publish '{}' v{}.".format(app_pkg,
                                                              self._version)
        print(message)
        if not self._ask_for_confirmation("\nDo you want to continue ?"):
            return
        print("")
        # check build_report
        if not self._check_build_report():
            return
        # get publishing hooks
        cache = self._get_hooks()
        self._pre_publishing_hook, self._post_publishing_hook = cache
        # execute pre_publishing_hook.py
        if not self._run_pre_publishing_hook():
            return
        # publish
        if not self._publish():
            return
        # execute post_publishing_hook.py
        if not self._run_post_publishing_hook():
            return
        print("\nSuccessfully published !")

    def _check_build_report(self):
        path = os.path.join(self._target, self._app_pkg,
                            "pyrustic_data", "build_report.json")
        if os.path.exists(path):
            return True
        text = ("\nMissing 'build_report.json' file in 'pyrustic_data' folder.",
                "Please build the project first with the command 'build'")
        print("\n".join(text))
        return False

    def _run_pre_publishing_hook(self):
        if not self._pre_publishing_hook:
            return True
        args = ["-m", self._pre_publishing_hook, self._target,
                self._app_pkg, self._version]
        p = subprocess.Popen([sys.executable, *args],
                             cwd=self._target)
        p.communicate()
        code = p.returncode
        if code == 0:
            print("pre_publishing_hook.py executed with success")
            return True
        else:
            print("Failed to execute pre_publishing_hook.py")
            return False

    def _run_post_publishing_hook(self):
        if not self._post_publishing_hook:
            return True
        args = ["-m", self._post_publishing_hook, self._target,
                self._app_pkg, self._version]
        p = subprocess.Popen([sys.executable, *args],
                             cwd=self._target)
        p.communicate()
        code = p.returncode
        print("")
        if code == 0:
            print("post_publishing_hook.py executed with success")
            return True
        else:
            print("Failed to execute pre_publishing_hook.py")
            return False

    def _get_hooks(self):
        dev_json = os.path.join(self._target,
                                self._app_pkg,
                                "pyrustic_data",
                                "dev.json")
        if not os.path.exists(dev_json):
            return None, None
        jasonix = Jasonix(dev_json)
        hooks_package = jasonix.data.get("hooking_pkg", None)
        if not hooks_package:
            return None, None
        pre_publishing_hook = "{}.pre_publishing_hook".format(hooks_package)
        post_publishing_hook = "{}.post_publishing_hook".format(hooks_package)
        return pre_publishing_hook, post_publishing_hook

    def _publish(self):
        publishing_json_path = os.path.join(self._target, self._app_pkg,
                                            "pyrustic_data",
                                            "publishing.json")
        if not os.path.exists(publishing_json_path):
            print("Missing 'publishing.json' in 'pyrustic_data' folder.\nPlease init your app.")
            return False
        jasonix = Jasonix(publishing_json_path)
        owner = jasonix.data.get("owner", None)
        repo = jasonix.data.get("repo", None)
        name = jasonix.data.get("name", None)
        tag_name = jasonix.data.get("tag_name", None)
        target_commitish = jasonix.data.get("target_commitish", None)
        description = jasonix.data.get("description", None)
        prerelease = jasonix.data.get("prerelease", None)
        draft = jasonix.data.get("draft", None)
        asset_path = jasonix.data.get("asset_path", None)
        asset_name = jasonix.data.get("asset_name", None)
        asset_label = jasonix.data.get("asset_label", None)
        if (not name or not tag_name
                or not asset_path or not owner or not repo):
            print("Missing mandatory elements in $APP_DIR/pyrustic_data/publishing.json")
            return False
        gurl = create_gurl()
        gurl.token = getpass.getpass("\nYour Github Token: ")
        print("\nProcessing...")
        catapult = Catapult(gurl, owner, repo)
        cache = catapult.publish(name, tag_name,target_commitish,
                         description, prerelease, draft,
                         asset_path, asset_name, asset_label)
        meta_code = cache["meta_code"]
        status_code = cache["status_code"]
        status_text = cache["status_text"]
        if meta_code == 0:
            return True
        if meta_code == 1:
            print("Failed to create release.\n{} {}".format(status_code,
                                                            status_text))
            return False
        if meta_code == 2:
            print("Failed to upload asset.\n{} {}".format(status_code,
                                                            status_text))
            return False
        print("Unknown error")
        return False

    def _get_version(self):
        setup_config_dict = setup_config(self._target)
        version = None
        if setup_config_dict:
            version = setup_config_dict["version"]
        return version

    def _ask_for_confirmation(self, message, default="y"):
        cache = "Y/n" if default == "y" else "y/N"
        user_input = input("{} ({}): ".format(message, cache))
        if not user_input:
            user_input = default
        if user_input.lower() == "y":
            return True
        return False



import tkinter as tk


def center_window(window, within=None):
    """ Center the window within another window (tk obj) or the screen (None)"""
    window.withdraw()
    window.update_idletasks()
    window.geometry("+0+0")
    window.update_idletasks()
    width = window.winfo_width()
    height = window.winfo_height()
    if within is None:
        x = (window.winfo_screenwidth() - width) // 2
        y = (window.winfo_screenheight() - height) // 2
    else:
        if isinstance(within, tk.Tk):
            pass
        elif isinstance(within, tk.Toplevel):
            pass
        else:
            within = within.winfo_toplevel()
        data = formal_geometry(within)
        x = ((data[0] // 2) + data[2]) - (width//2)
        y = ((data[1] // 2) + data[3]) - (height//2)
    if (window.winfo_screenwidth() - x) < width:
        x = window.winfo_screenwidth() - width
    if (window.winfo_screenheight() - y) < height:
        y = window.winfo_screenheight() - height
    window.geometry("+{}+{}".format(x, y))
    window.deiconify()


def align_window(window, under=None):
    window.withdraw()
    window.update_idletasks()
    width = window.winfo_width()
    height = window.winfo_height()
    under_x = under.winfo_rootx()
    under_y = under.winfo_rooty()
    x = under_x
    y = under_y + under.winfo_height()
    x = abs(x)
    y = abs(y)
    if window.winfo_screenwidth() - x < width:
        x = window.winfo_screenwidth() - width
    if window.winfo_screenheight() - y < height:
        y = window.winfo_screenheight() - height
    # align
    window.geometry("+{}+{}".format(x, y))
    window.deiconify()


def dialog_effect(window):
    window.transient(window.master)
    window.lift()
    window.grab_set()
    window.focus_set()


def formal_geometry(window):
    width = window.winfo_width()
    height = window.winfo_height()
    coord_x = window.winfo_x()
    coord_y = window.winfo_y()
    return width, height, coord_x, coord_y


def merge_cnfs(main_cnfs, extra_cnfs, components=None):
    main_cnfs = {} if not main_cnfs else main_cnfs
    extra_cnfs = {} if not extra_cnfs else extra_cnfs
    components = {} if not components else components
    cnfs = {}
    # components
    for item in components:
        cnfs[item] = {}
    # parse extra_cnf
    for key, val in extra_cnfs.items():
        cnfs[key] = val
    # parse main_cnf
    for key, val in main_cnfs.items():
        if val is not None:
            cnfs[key] = val
    return cnfs



from pyrustic.manager.misc import funcs
import os
import os.path
import pkgutil


class AddHandler:
    """
    Description
    -----------
    Use this command to add a file, a package or a regular folder
    to the Target.
    Pyrustic Manager will assume that files with extension ".py"
    are modules.
    Some convenient lines of code may be automatically inserted
    into modules based on the package name.

    Usage
    -----
    - Description: Add a file
    - Command: add <destination> <file.ext>

    - Description: Add multiple files
    - Command: add <destination> <file_1.ext> <file_2.ext>

    - Description: Add a module
    - Command: add <destination> <file.py>

    - Description: Add a package
    - Command: add <my.pack.age>

    - Description: Add a file to the Target's APP_DIR
    - Command: add ./ <file.ext>

    - Description: Add a module to the Target's APP_DIR
    - Command: add . <file.py>

    Note: The destination is either a relative path to the
    Target's APP_DIR, or a package name.
    Also, only SLASH is allowed in the path as separator.

    Example
    -------
    - Description: Add a module
    - Preliminary: Assume you want to add "my_view.py"
    to the package 'demo.view'
    - Command: add demo.view my_view.py

    - Description: Add a package
    - Preliminary: Assume you want to add 'my.new.package'
    - Command: add my.new.package

    Note: Please use simple or double quotes as delimiters if a
    string contains space
    """
    def __init__(self, target,
                 app_pkg, args):
        self._target = target
        self._app_pkg = app_pkg
        self._args = args
        self._process(target, app_pkg, args)

    def _process(self, target, app_pkg, args):
        if not target:
            self._print_catalog("missing_target")
            return
        if not app_pkg:
            print("Please init the project first. Check 'help init'.")
            return
        if not args:
            self._print_catalog("incomplete")
            return
        # package or folder
        package_or_folder_name = args[0]
        if (package_or_folder_name == "."
                or ("/" not in package_or_folder_name
                and "\\" not in package_or_folder_name)):
            self._process_package_and_files(target, args)
        else:
            self._process_folder_and_files(target, args)

    def _process_package_and_files(self, target, args):
        package_name = args[0]
        if len(args) > 1:
            if not self._handle_package_part(target, package_name, verbose=False):
                return
            self._add_files_into_package(target, package_name, args[1:])
        else:
            self._handle_package_part(target, package_name)

    def _process_folder_and_files(self, target, args):
        folder = args[0]
        if len(args) > 1:
            if not self._handle_folder_part(target, folder, verbose=False):
                return
            self._add_files_into_folder(target, folder, args[1:])
        else:
            self._handle_folder_part(target, folder)

    def _handle_package_part(self, target, package_name, verbose=True):
        path = funcs.package_name_to_path(target, package_name)
        if os.path.isdir(path):
            if verbose:
                self._print_catalog("package_already_exists")
            return True
        self._print_catalog("package_not_exists")
        self._print_catalog("building_package", package_name=package_name)
        path = funcs.build_package(target, package_name)
        if path is None:
            self._print_catalog("failed_to_build_package")
            return False
        else:
            self._print_catalog("package_built")
            return True

    def _add_files_into_package(self, target, package_name, files):
        for file in files:
            # Is module
            if file.endswith(".py"):
                if not self._install_module(target, package_name, file):
                    return False
            # Is simple file
            else:
                if not self._install_file(target, package_name, file):
                    return False
        return True

    def _handle_folder_part(self, target, folder, verbose=True):
        path = os.path.join(target, folder)
        path = os.path.normpath(path)
        if os.path.isdir(path):
            if verbose:
                self._print_catalog("folder_already_exists")
            return True
        self._print_catalog("folder_not_exists")
        self._print_catalog("creating_folder", folder=folder)
        try:
            os.makedirs(path)
        except Exception as e:
            self._print_catalog("failed_to_create_folder")
            return False
        else:
            self._print_catalog("folder_created")
            return True

    def _add_files_into_folder(self, target, folder, files):
        path = os.path.join(target, folder)
        path = os.path.normpath(path)
        for filename in files:
            path = os.path.join(path, filename)
            if os.path.exists(path):
                self._print_catalog("file_exists", filename=filename)
                continue
            try:
                with open(path, "w") as file:
                    pass
            except Exception as e:
                self._print_catalog("file_creation_failed", filename=filename)
                return False
            self._print_catalog("file_created", filename=filename)
        return True

    def _install_module(self, target, package_name, filename):
        # creating file
        class_name = funcs.module_name_to_class(filename)
        path = funcs.package_name_to_path(target, package_name)
        path = os.path.join(path, filename)
        if os.path.exists(path):
            self._print_catalog("module_exists", module=filename)
            return True
        resource_prefix = "manager/template/"
        if not self._app_pkg:
            self._print_catalog("module_created", module=filename)
            return True
        with open(path, "wb") as file:
            if package_name.startswith("{}.view".format(self._app_pkg)):
                resource = resource_prefix + "view_template.txt"
                data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
                data = data.format(app_pkg=self._app_pkg,
                                   class_name=class_name)
                file.write(data)
            else:
                print("HEEEREE !!!!")
                resource = resource_prefix + "module_template.txt"
                data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
                data = data.format(app_pkg=self._app_pkg,
                                   class_name=class_name)
                file.write(data)
            self._print_catalog("module_created", module=filename)
        return True

    def _install_file(self, target, package_name, filename):
        path = funcs.package_name_to_path(target, package_name)
        path = os.path.join(path, filename)
        if os.path.exists(path):
            self._print_catalog("file_exists", filename=filename)
            return True
        try:
            with open(path, "w") as file:
                pass
        except Exception as e:
            self._print_catalog("file_creation_failed", filename=filename)
            return False
        self._print_catalog("file_created", filename=filename)
        return True

    def _create_file(self):
        pass

    def _get_data_from_template(self, path):
        data = ""
        with open(path, "r") as file:
            data = file.read()
        return data

    def _print_catalog(self, item, **kwargs):
        message = ""
        if item == "missing_target":
            message = "Please link a Target first. Check 'help target'."
        elif item == "incomplete":
            message = "Incomplete command. Check 'help add'."
        elif item == "package_already_exists":
            message = "This package already exists !"
        elif item == "package_not_exists":
            message = "This package doesn't exist yet."
        elif item == "building_package":
            message = "Building package '{}'".format(kwargs["package_name"])
        elif item == "package_built":
            message = "Package successfully built."
        elif item == "failed_to_build_package":
            message = "Failed to finish the package building."
        elif item == "file_created":
            message = "File '{}' successfully created.".format(kwargs["filename"])
        elif item == "file_creation_failed":
            message = "Failed to create file '{}'.".format(kwargs["filename"])
        elif item == "file_exists":
            message = "The file '{}' already exists.".format(kwargs["filename"])
        elif item == "module_exists":
            message = "The module '{}' already exists.".format(kwargs["module"])
        elif item == "module_created":
            message = "Module '{}' successfully created.".format(kwargs["module"])
        elif item == "folder_already_exists":
            message = "This folder already exists !"
        elif item == "folder_not_exists":
            message = "This folder doesn't exist yet."
        elif item == "creating_folder":
            message = "Creating folder '{}'".format(kwargs["folder"])
        elif item == "folder_created":
            message = "Folder successfully created."
        elif item == "failed_to_create_folder":
            message = "Failed to create the folder."
        print(message)



import os.path
from pyrustic import dist
from pyrustic.manager.misc import funcs


class TargetHandler:
    """
    Description
    -----------
    Use this command to check the currently linked Target.

    Usage
    -----
    - Description: Check the currently linked Target
    - Command: target

    Note: This command also shows the version of your project
    if the data is available.
    """
    def __init__(self, target,
                 app_pkg,
                 args):
        self._target = target
        self._app_pkg = app_pkg
        self._process(target, app_pkg, args)

    @property
    def target(self):
        return self._target

    def _process(self, target, app_pkg, args):
        if not target:
            print("None")
            return
        # args are present: invalid command
        if args:
            print("Wrong usage of this command")
            return
        #
        data = funcs.check_project_state(target)
        print("[{}] {}".format(os.path.basename(target), self._target))
        if data == 0:
            print("Version: {}".format(dist(app_pkg)["version"]))
        elif data == 1:
            print("Not yet initialized project (check 'help init')")
        elif data == 2:
            print("Not yet installed project (think about: 'pip install -e .')")



import tkinter as tk
from pyrustic import tkmisc
from pyrustic.exception import PyrusticException


# Constants
NEW = "new"
BUILT = "built"
DISPLAYED = "displayed"
DESTROYED = "destroyed"


class View:
    """
    Subclass this if you are going to create a view.

    Lifecycle of a view:
        1- you instantiate the view
        2- '__init__()' is implicitly called
        3- you call the method '.build()'
        4- '_on_build()' is implicitly called
        5- '_on_display()' is implicitly called once the widget is visible
        6- '_on_destroy()' is implicitly called when the widget is destroyed/closed

    The rules to create your view is simple:
    - You need to subclass Viewable.
    - You need to implement the methods '_on_build()', and optionally
        implement '_on_display()' and '_on_destroy()'.
    - You need to set an instance variable '_body' with either a tk.Frame or tk.Toplevel
        in the method '_on_build()'
    That's all ! Of course, when you are ready to use the view, just call the 'build()' method.
    Calling the 'build()' method will return the body of the view. The one that you assigned
    to the instance variable '_body'. The same body can be retrieved with the property 'body'.
    The 'build()' method should be called once. Calling it more than once will still return
    the body object, but the view won't be built again.
    You can't re-build your same view instance after destroying its body.
    You can destroy the body directly, by calling the conventional tkinter destruction method
     on the view's body. But it's recommended to destroy the view by calling the view's method
     'destroy()' inherited from the class Viewable.
    The difference between these two ways of destruction is that when u call the Viewable's
     'destroy()' method, the method '_on_destroy()' will be called BEFORE the effective
     destruction of the body. If u call directly 'destroy' conventionally on the tkinter
     object (the body), the method '_on_destroy()' will be called AFTER the beginning
      of destruction of the body.

      By the way, you can use convenience methods "build_pack", "build_grid", "build_place"
      to build and pack/grid/place your widget in the master !!
      Use "build_wait" for toplevels if you want the app to wait till the window closes
    """

    def __init__(self):
        self.__master = None
        self.__state = 0
        self.__built = False
        self.__bind_id = None
        self.__destroyed = False
        self._body = None

    # ==============================================
    #                 PROPERTIES
    # ==============================================

    @property
    def body(self):
        """
        Get the body of this view.
        """
        return self._body

    @property
    def state(self):
        """ Return the current state of the Viewable instance.
        States are integers, you can use these constants:
            - pyrustic.view.NEW: the state just after instantiation;
            - pyrustic.view.BUILT: the state after the call of on_body
            - pyrustic.view.DISPLAYED: the state after the call of on_display
            - pyrustic.view.DESTROYED: the state after the call of on_destroy
        """
        return self.__state
    # ==============================================
    #                 PUBLIC METHODS
    # ==============================================

    def build(self):
        """
        Build the view. Return the body
        """
        return self.__build()

    def build_pack(self, cnf=None, **kwargs):
        cnf = {} if not cnf else cnf
        body = self.__build()
        self._check_missing_body(body)
        body.pack(cnf=cnf, **kwargs)

    def build_grid(self, cnf=None, **kwargs):
        cnf = {} if not cnf else cnf
        body = self.__build()
        self._check_missing_body(body)
        body.grid(cnf=cnf, **kwargs)

    def build_place(self, cnf=None, **kwargs):
        cnf = {} if not cnf else cnf
        body = self.__build()
        self._check_missing_body(body)
        body.place(cnf=cnf, **kwargs)

    def build_wait(self):
        """
        Build the view. Return the body
        """
        body = self.__build()
        body.wait_window(body)

    def destroy(self):
        """
        Destroy the body of this view
        """
        self.__build()
        self.__exec_on_destroy()


    # ==============================================
    #               METHODS TO IMPLEMENT
    # ==============================================

    def _on_build(self):
        """
        Put here the code that build the body of this view.
        The body is either a tk.Frame or a tk.Toplevel instance.
        """
        pass

    def _on_display(self):
        """
        Put here the code that will be executed once the body is visible.
        """
        pass

    def _on_destroy(self):
        """
        Put here the code that will be executed as clean-up.
        """
        pass

    def _toplevel_geometry(self):
        """
        If the body of this view is a toplevel and
        you need to change the geometry of this toplevel,
        override this method !
        """
        tkmisc.center_window(self.body, self.__master.winfo_toplevel())
        tkmisc.dialog_effect(self.body)

    # ==============================================
    #                 INTERNAL METHODS
    # ==============================================

    def __build(self):
        if self.__built:
            return self.body
        self._on_build()
        self.__built = True
        self.__state = BUILT
        if not self.body:
            return
        try:
            self.__master = self.body.master
        except Exception:
            pass
        is_toplevel = isinstance(self.body, tk.Toplevel)
        #is_frame = isinstance(self.body, tk.Frame)
        if is_toplevel:
            #self.body.protocol("WM_DELETE_WINDOW", self.__exec_on_destroy)
            self._toplevel_geometry()
        self.__bind_destroy_event()
        if self.body.winfo_viewable():
            self.__exec_on_display()
        else:
            self.__bind_id = self.body.bind("<Map>",
                                            self.__exec_on_display,
                                            "+")
        return self.body

    def __exec_on_display(self, event=None):
        self._on_display()
        self.__state = DISPLAYED
        if self.__bind_id is not None:
            self.body.unbind("<Map>", self.__bind_id)

    def __bind_destroy_event(self):
        command = (lambda event,
                          widget=self.body,
                          callback=self.__exec_on_destroy:
                   callback() if event.widget is widget else None)
        self.body.bind("<Destroy>", command, "+")

    def __exec_on_destroy(self):
        if not self.__built or self.__destroyed:
            return
        if not self.body:
            return
        window_manager = self.body.winfo_manager()
        """
        # Hide the window first to avoid the visual
        # iterative (slow) destruction
        # of each child
        #
        if window_manager == "wm":
            if self.body.winfo_ismapped():
                self.body.withdraw()
        elif window_manager == "grid":
            if self.body.winfo_ismapped():
                self.body.grid_forget()
        elif window_manager == "pack":
            if self.body.winfo_ismapped():
                self.body.pack_forget()
        elif window_manager == "place":
            if self.body.winfo_ismapped():
                self.body.place_forget()
        """
        try:
            self.body.destroy()
        except Exception as e:
            pass
        self.__destroyed = True
        self._on_destroy()
        self.__state = DESTROYED
        try:
            if self.__master.focus_get() is None:
                self.__master.winfo_toplevel().focus_lastfor().focus_force()
        except Exception as e:
            pass

    def _check_missing_body(self, body):
        if not body:
            raise PyrusticException("Missing body")


class CustomView(View):
    def __init__(self, body=None, on_build=None,
                 on_display=None,
                 on_destroy=None,
                 toplevel_geometry=None):
        super().__init__()
        self._body = body
        if on_build:
            self._on_build = on_build
        if on_display:
            self._on_display = on_display
        if on_destroy:
            self._on_destroy = on_destroy
        if toplevel_geometry:
            self._toplevel_geometry = toplevel_geometry



import sys
import copy
import platform
import pkgutil
import tkinter as tk
import pyrustic
from pyrustic.view import View, CustomView
from pyrustic import tkmisc
from pyrustic.jasonix import Jasonix
from pyrustic.exception import PyrusticAppException
from pyrustic.private.enhance_tk import EnhanceTk


class App:
    """
    Pyrustic Framework's entry point.
    This class should be instantiated inside the file "main.py".
    """
    def __init__(self, package):
        """
        Create an App instance.
        package: the name of the package in which the caller is. Use __package__.

        It's recommended to don't write any code above this instantiation.
        """
        self._package = package
        self._is_running = False
        self._restartable = False
        self._root = tk.Tk()
        self._theme = None
        self._theme_cache = None
        self._view = None
        self._view_cache = None
        self._center_window = False
        self._config = None
        self._gui_config = None
        self._exit_handler = None
        self._setup()

    # ============================================
    #                PROPERTIES
    # ============================================
    @property
    def root(self):
        """
        Get the main tk root
        """
        return self._root

    @property
    def installed(self):
        data = pyrustic.dist(self._package)
        return True if data else False

    @property
    def config(self):
        return copy.deepcopy(self._config)

    @config.setter
    def config(self, val):
        """ val is dict, path or file-like object"""
        jasonix = Jasonix(val)
        self._config = jasonix.data
        if self._config:
            self._gui_config = self._config.get("gui", self._gui_config)

    @property
    def gui_config(self):
        """
        Setter et Getter
        """
        return copy.deepcopy(self._gui_config)

    @gui_config.setter
    def gui_config(self, val):
        jasonix = Jasonix(val)
        self._gui_config = jasonix.data
        if self._gui_config:
            self._config["gui"] = self._gui_config

    @property
    def theme(self):
        """
        Get the theme object
        For more information about what a theme is:
        - check 'pyrustic.theme.Theme';
        - then check the beautiful theme 'pyrustic.theme.cyberpunk'
        """
        return self._theme

    @theme.setter
    def theme(self, val):
        """
        Set the theme object.
        If you set None, it will invalidate the previous theme.
        Don't forget to call the method "restart()" or "start()" to apply the change !
        Remember that "start()" should be called only once !
        For more information about what a theme is:
        - check "pyrustic.theme.Theme";
        - then check the beautiful theme "pyrustic.theme.cyberpunk"
        """
        self._theme_cache = val

    @property
    def view(self):
        """
        Get the view object.
        A view should implement "pyrustic.viewable.Viewable"
        """
        return self._view

    @view.setter
    def view(self, val):
        """
        Set a view object.
        If you set None, the previous view will be destroyed.
        The new view will destroy the previous one if there are a previous one.
        VAL can be a tkinter object or a callable (if u plan to REFRESH the app)
        """
        self._view_cache = val

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, val):
        self._body = val

    @property
    def exit_handler(self):
        return self._exit_handler

    @exit_handler.setter
    def exit_handler(self, val):
        self._exit_handler = val

    # ============================================
    #               PUBLIC METHODS
    # ============================================
    def start(self):
        """
        Call this method to start the app.
        It should be called once and put on the last line of the file.
        """
        if self._is_running:
            message = "This method shouldn't be called twice. Please use 'restart' instead"
            raise PyrusticAppException(message)
        self._is_running = True
        self._build()
        # main loop
        try:
            self._root.mainloop()
        except KeyboardInterrupt:
            pass

    def refresh(self):
        """
        Call this method to refresh the app.
        If you have submitted a new view or a new theme,
        the previous view or theme will be removed and the new one installed
        """
        if not self._is_running:
            return False
        if not self._view_cache or not callable(self._view_cache):
            return False
        if self._theme_cache:
            self._apply_theme(self._theme_cache)
            self._theme_cache = None
        if self._view:
            self._view.destroy()
        self._body = None
        return self._install_view(self._view_cache,
                                  is_refresh=True)

    def exit(self):
        """
        Exit, simply ;-)
        Depending on your config file, the application will close quickly or not.
        A quick exit will ignore the lifecycle of a Viewable (pyrustic.viewable).
        In others words, '_on_destroy()' methods won't be called.
        Exit quickly if you don't care clean-up but want the app to close as fast as possible.
        """
        self._on_exit()

    def maximize(self):
        """
        Maximize the window
        """
        system = platform.system()
        if system == "Linux":
            self._root.attributes("-zoomed", True)
        else:  # for "Darwin" (OSX) and "Window"
            self._root.state("zoomed")

    def center(self):
        """
        Center the window
        """
        self._center_window = True

    # ============================================
    #               PRIVATE METHODS
    # ============================================

    def _setup(self):
        # gui_config
        self._set_config()
        # set default title
        self._set_default_title()

    def _build(self):
        # bind self._on_exit
        self._root.protocol("WM_DELETE_WINDOW", self._on_exit)
        handler = (lambda event,
                          root=self._root:
                   self._on_exit() if event.widget is root else None)
        self._root.bind("<Destroy>", handler)
        EnhanceTk(self._root)
        # apply config, set theme then install view
        self._apply_config()
        self._apply_theme(self._theme_cache)
        self._install_view(self._view_cache)

    def _set_config(self):
        self._set_gui_config()

    def _set_gui_config(self):
        gui_config_json = None
        gui_config_json_resource = "pyrustic_data/gui.json"
        default_gui_json_resource = \
            "manager/default_json/pyrustic_data/gui_default.json"
        if self._package:
            try:
                gui_config_json = pkgutil.get_data(self._package,
                                                   gui_config_json_resource)
            except Exception as e:
                pass
        if not gui_config_json:
            gui_config_json = pkgutil.get_data(__package__,
                                               default_gui_json_resource)
        jasonix = Jasonix(gui_config_json)
        self._gui_config = jasonix.data
        self._config = {"gui": self._gui_config}

    def _set_default_title(self):
        name = self._package
        if not self._package or "." in name:
            name = "Application"
        title = "{} | built with Pyrustic".format(name)
        self._root.title(title)

    def _apply_config(self):
        self._apply_gui_config()

    def _apply_gui_config(self):
        # app geometry
        if not self._gui_config["ignore_geometry"]:
            self._root.geometry(self._gui_config["root_geometry"])
        # background
        background_color = self._gui_config["root_background"]
        self._root.config(background=background_color)
        # resizable width and height
        resizable_width = self._gui_config["resizable_width"]
        resizable_height = self._gui_config["resizable_height"]
        self._root.resizable(width=resizable_width, height=resizable_height)
        # maximize screen
        if self._gui_config["maximize_window"]:
            self.maximize()

    def _apply_theme(self, theme):
        if self._theme:
            self._root.option_clear()
        self._theme = theme
        if not self._gui_config["allow_theme"]:
            return
        if not self._theme:
            return
        self._theme.target(self._root)

    def _install_view(self, view, is_refresh=False):
        self._view = self._get_view(view)
        self._view.build()
        if not self._view.build():
            return False
        body = self._view.body
        if isinstance(body, tk.Frame):
            self._view.body.pack(in_=self._root,
                                 expand=1, fill=tk.BOTH)
        elif isinstance(body, tk.Toplevel):
            pass
        else:
            self._view.body.pack(in_=self._root)
        # center
        if not is_refresh and self._center_window:
            tkmisc.center_window(self._root)
        return True

    def _get_view(self, view):
        if callable(view):
            view = view()
        if isinstance(view, View):
            return view
        if isinstance(view, type) and issubclass(view, View):
            return view()
        if view is None:
            view = tk.Frame(self._root,
                            bg="black",
                            width=350,
                            height=200)
        return CustomView(body=view)

    def _on_exit(self):
        if self._exit_handler:
            val = self._exit_handler()
            if not val:
                return
        if self._view:
            if self._view.body:
                pass
                #self._root.iconify()
                #self._root.withdraw()
            self._view.destroy()
            self._view = None
        if self._root:
            self._root.destroy()
            self._root = None
        sys.exit()



import tkinter as tk
from pyrustic import tkmisc
from pyrustic import widget
from pyrustic.widget.scrollbox import Scrollbox
from pyrustic.view import View
from pyrustic.tkmisc import merge_cnfs


# select button flavor
CHECK = "check"  # for checkbutton
RADIO = "radio"  # for radiobutton


# Components
LABEL_HEADER = "label_header"
SCROLLBOX = "scrollbox"
LABEL_MESSAGE = "label_message"
FRAME_PANE = "frame_pane"
FRAME_FOOTER = "frame_footer"
BUTTON_CONTINUE = "button_continue"
BUTTON_CANCEL = "button_cancel"
RADIOBUTTONS = "radiobuttons"
CHECKBUTTONS = "checkbuttons"


class Choice(widget.Toplevel):
    """
    Choice is a dialog box to make the user select some items among others.
    The Choice could be implemented with either radiobuttons or checkbuttons.

    Example:

        import tkinter as tk
        from pyrustic.widget.choice import Choice

        def my_handler(result):
            print(result)

        root = tk.Tk()
        my_items = ("first", "second", "third")
        choice = Choice(root, title="Choice", header="Make a choice",
                        items=my_items, handler=my_handler)
        choice.build()
        root.mainloop()

    """

    def __init__(self,
                 master=None,
                 title=None,
                 header=None,
                 message=None,
                 items=None,
                 selected=None,
                 flavor="radio",
                 handler=None,
                 geometry=None,
                 cnfs=None):
        """
        PARAMETERS:

        - master: widget parent. Example: an instance of tk.Frame

        - title: title of dialog box

        - header: the text to show as header

        - message: the text to show as message

        - use_scrollbox: bool, set it to True to make the Dialog scrollable

        - items: a sequence of strings. Example: ("banana", "apple").

        - selected: a sequence of indexes to indicate default selection.
        Set it to None if u don't need it.

        - flavor: it could be either RADIO or CHECK
        for respectively radiobutton and checkbutton

        - handler: a callback to be executed immediately
        after closing the dialog box.
        The callback should allow one parameter, the result:

            - If the flavor is RADIO,
             then, result is a tuple like: (the selected index, item string).

            - If the flavor is CHECK,
             then, result is a sequence of tuples.
             Each tuple is like: (integer, item string),
             with integer being 1 if the button has been clicked, else 0.

        - geometry: str, as the dialog box is a toplevel (BODY),
         you can edit its geometry. Example: "500x300"

        - options: dictionary of widgets options
            The widgets keys are: BODY, LABEL_HEADER, SCROLLBOX, LABEL_MESSAGE,
            FRAME_PANE, FRAME_FOOTER, BUTTON_CONTINUE, BUTTON_CANCEL,
            RADIOBUTTONS, CHECKBUTTONS.

            Example: Assume that you want to set the LABEL_MESSAGE's background to black
            and the BODY's background to red:
                options = { BODY: {"background": "red"},
                            LABEL_MESSAGE: {"background": "black"} }

        """
        self.__cnfs = merge_cnfs(None, cnfs, components=("body",
                                LABEL_HEADER, SCROLLBOX, LABEL_MESSAGE,
                                FRAME_PANE, FRAME_FOOTER, BUTTON_CONTINUE,
                                BUTTON_CANCEL, RADIOBUTTONS, CHECKBUTTONS))
        super().__init__(master=master,
                         class_="Choice",
                         cnf=self.__cnfs["body"],
                         on_build=self.__on_build,
                         on_display=self.__on_display,
                         on_destroy=self.__on_destroy,
                         toplevel_geometry=self.__toplevel_geometry)
        self.__title = title
        self.__header = header
        self.__message = message
        self.__items = [] if not items else items
        self.__selected = selected
        self.__flavor = flavor
        self.__handler = handler
        self.__geometry = geometry
        #
        self.__result = None
        self.__closing_context = "close"
        self.__components = dict()
        self.__label_header = None
        self.__label_message = None
        self.__pane = None
        self.__footer = None
        self.__buttons = None
        self.__intvar = tk.IntVar()
        self.__intvars = []
        # components
        self.__components = {}
        # build
        self.__view = self.build()

    # ======================================
    #            PROPERTIES
    # ======================================

    @property
    def header(self):
        return self.__header

    @property
    def message(self):
        return self.__message

    @property
    def items(self):
        return self.__items.copy()

    @property
    def selected(self):
        """
        - If the flavor is RADIO,
             then, result is a tuple like: (the selected index, item string).
             Example: 3 items, the second has been selected:
                result = (1, "Item at index 1")

        - If the flavor is CHECK,
         then, result is a sequence of tuples, each positioned in
         the sequence according to its index number.
         Each tuple is like: (integer, item string),
         with integer being 1 if the button has been clicked, else 0.
         Example: 3 items, only the last 2 are checked:
            result = ( (0, "item 1"), (1, "item 2"), (1, "item 3") )
        """
        return self.__result

    @property
    def flavor(self):
        return self.__flavor

    @property
    def handler(self):
        return self.__handler

    @property
    def components(self):
        """
        Get the components (widgets instances) used to build this dialog.

        This property returns a dict. The keys are:
            BODY, LABEL_HEADER, SCROLLBOX, LABEL_MESSAGE,
            FRAME_PANE, FRAME_FOOTER, BUTTON_CONTINUE, BUTTON_CANCEL,
            RADIOBUTTONS, CHECKBUTTONS.

        Warning: radiobuttons and checkbuttons are sequences of widgets positioned
        in the sequence according to the index.

        Another Warning: check the presence of key before usage.
        """
        return self.__components

    # ======================================
    #            INTERNAL
    # ======================================
    def __on_build(self):
        self.title(self.__title)
        self.resizable(0, 0)
        #
        if self.__geometry:
            self.geometry(self.__geometry)
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=0)
        self.rowconfigure(1, weight=0)
        self.rowconfigure(2, weight=2, uniform="a")
        self.rowconfigure(3, weight=0, uniform="a")
        # == Set Header
        if self.__header:
            label_header = tk.Label(self,
                                    name=LABEL_HEADER,
                                    text=self.__header,
                                    justify=tk.LEFT,
                                    anchor="w",
                                    cnf=self.__cnfs[LABEL_HEADER])
            self.__components[LABEL_HEADER] = label_header
            label_header.grid(row=0, column=0, sticky="w",
                                    padx=(5, 5), pady=(5, 5))
            label_header.config(text=self.__header)
        # == Set Message
        if self.__message:
            label_message = tk.Label(self,
                                     name=LABEL_MESSAGE,
                                     text=self.__message,
                                     justify=tk.LEFT,
                                     anchor="w",
                                     cnf=self.__cnfs[LABEL_MESSAGE])
            self.__components[LABEL_MESSAGE] = label_message
            label_message.grid(row=1, column=0, sticky="w",
                               padx=(5, 5), pady=(0, 5))
        # == Scrollbox
        scrollbox = Scrollbox(self, orient="vertical",
                              cnfs=self.__cnfs[SCROLLBOX])
        self.__components[SCROLLBOX] = scrollbox
        scrollbox.grid(row=2, column=0, sticky="nswe",
                             padx=5)
        # == Footer
        self.__footer = tk.Frame(self,
                                 name=FRAME_FOOTER,
                                 cnf=self.__cnfs[FRAME_FOOTER])
        self.__components[FRAME_FOOTER] = self.__footer
        self.__footer.grid(row=3, column=0, sticky="swe", pady=(30, 0))
        #
        button_continue = tk.Button(self.__footer, name=BUTTON_CONTINUE,
                                    text="Continue",
                                    command=self.__on_click_continue,
                                    cnf=self.__cnfs[BUTTON_CONTINUE])
        self.__components[BUTTON_CONTINUE] = button_continue
        button_continue.pack(side=tk.RIGHT, padx=2, pady=2)
        #
        button_cancel = tk.Button(self.__footer, name=BUTTON_CANCEL,
                                  text="Cancel",
                                  command=self.__on_click_cancel,
                                  cnf=self.__cnfs[BUTTON_CANCEL])
        self.__components[BUTTON_CANCEL] = button_cancel
        button_cancel.pack(side=tk.RIGHT, padx=(2, 0), pady=2)
        # install and populate check/radio buttons
        key = RADIOBUTTONS if self.__flavor == "radio" else CHECKBUTTONS
        self.__components[key] = []
        cache = None
        for i, choice in enumerate(self.__items):
            if not self.__flavor or self.__flavor not in ("radio", "check"):
                break
            if self.__flavor == "radio":
                cache = tk.Radiobutton(scrollbox.box,
                                       variable=self.__intvar,
                                       text=choice, value=i,
                                       cnf=self.__cnfs[RADIOBUTTONS])
                self.__components[RADIOBUTTONS].append(cache)
            elif self.__flavor == "check":
                tk_var = tk.IntVar()
                self.__intvars.append(tk_var)
                cache = tk.Checkbutton(scrollbox.box,
                                       variable=tk_var,
                                       onvalue=1, offvalue=0,
                                       text=choice,
                                       cnf=self.__cnfs[CHECKBUTTONS])
                self.__components[CHECKBUTTONS].append(cache)
            if cache:
                cache.pack(anchor="w", expand=1)

    def __on_display(self):
        # fill selected items
        if self.__flavor == RADIO and self.__selected is not None:
            if isinstance(self.__selected, int) and self.__selected >= 0:
                self.__intvar.set(self.__selected)
        elif self.__flavor == CHECK and self.__selected is not None:
            if isinstance(self.__selected, tuple):
                for i in self.__selected:
                    try:
                        self.__intvars[i].set(1)
                    except IndexError:
                        pass
            elif isinstance(self.__selected, int):
                self.__intvars[self.__selected].set(1)

    def __on_destroy(self):
        if self.__closing_context == "continue":
            self.__result = self.__get_result()
        if self.__handler:
            self.__handler(self.__result)

    def __toplevel_geometry(self):
        tkmisc.center_window(self, within=self.master.winfo_toplevel())
        tkmisc.dialog_effect(self)

    def __on_click_continue(self):
        self.__closing_context = "continue"
        self.destroy()

    def __on_click_cancel(self):
        self.__closing_context = "cancel"
        self.destroy()

    def __get_result(self):
        result = None
        if self.__flavor == "radio":
            index = self.__intvar.get()
            result = (index, self.__items[index])
        elif self.__flavor == "check":
            cache = []
            for i, intvar in enumerate(self.__intvars):
                intvar_index = intvar.get()
                cache.append((intvar_index, self.__items[i]))
            result = tuple(cache)
        return result


class _ChoiceTest(View):
    def __init__(self, root):
        super().__init__()
        self._root = root
        self._body = None

    def _on_build(self):
        self._body = tk.Frame(self._root)
        btn_launch_check_choice = tk.Button(self._body,
                                            text="Launch checkbutton choice",
                                            command=self._on_click_btn_check)
        btn_launch_check_choice.pack(side=tk.LEFT, anchor="nw")
        btn_launch_radio_choice = tk.Button(self._body,
                                            text="Launch radiobutton choice",
                                            command=self._on_click_btn_radio)
        btn_launch_radio_choice.pack(side=tk.LEFT, anchor="nw")


    def _on_display(self):
        pass

    def _on_destroy(self):
        pass

    def _on_click_btn_check(self):
        Choice(self._body, title="Title", header="header", flavor="check",
                        message="message",
                        items=["first", "second", "third"],
                        selected=1,
               handler=self._choice_handler)

    def _on_click_btn_radio(self):
        tests = ("test "*10).split()
        Choice(self._root,
               title="Title", header="header", flavor="radio",
               message="message",
               items=["first", "second", "third", *tests],
               selected=1, handler=self._choice_handler)

    def _choice_handler(self, data):
        print("Choice: {}".format(data))


if __name__ == "__main__":
    root = tk.Tk()
    root.geometry("500x300+0+0")
    choice_test = _ChoiceTest(root)
    choice_test.build_pack()
    root.mainloop()



import tkinter as tk
from tkinter import filedialog
from pyrustic import widget
from pyrustic.tkmisc import merge_cnfs
from pyrustic.view import View


ENTRY = "entry"
BUTTON = "button"
DIALOG = "dialog"


class Pathentry(widget.Frame):
    """
    """
    def __init__(self,
                 master=None,
                 browse="file",
                 width=17,
                 title=None,
                 initialdir=None,
                 cnfs=None):
        """
        - master: widget parent. Example: an instance of tk.Frame

        """
        self.__cnfs = merge_cnfs({ENTRY: {"width": width}}, cnfs,
                                 components=("body", ENTRY, BUTTON, DIALOG))
        super().__init__(master=master,
                         class_="Pathentry",
                         cnf=self.__cnfs["body"],
                         on_build=self.__on_build,
                         on_display=self.__on_display,
                         on_destroy=self.__on_destroy)
        self.__browse = browse
        self.__title = title
        self.__initialdir = initialdir
        self.__entry = None
        self.__button = None
        self.__components = {}
        self.__string_var = tk.StringVar(value="")
        # build
        self.__view = self.build()
    # ==============================================
    #                   PROPERTIES
    # ==============================================
    @property
    def components(self):
        """
        """
        return self.__components

    @property
    def string_var(self):
        return self.__string_var

    @property
    def path(self):
        return self.__path

    @path.setter
    def path(self, val):
        self.__path = val

    def __on_build(self):
        self.__entry = tk.Entry(self, textvariable=self.__string_var,
                                cnf=self.__cnfs[ENTRY])
        self.__entry.pack(side=tk.LEFT, pady=0, fill=tk.X, expand=1)
        self.__components["entry"] = self.__entry
        self.__button = tk.Button(self, text="...",
                                  command=self.__on_click_button,
                                  cnf=self.__cnfs[BUTTON])
        self.__button.pack(side=tk.LEFT, padx=(2, 0), fill=tk.Y)
        self.__components["button"] = self.__button

    def __on_display(self):
        pass

    def __on_destroy(self):
        pass

    def __on_click_button(self):
        if self.__browse == "file":
            try:
                filename = filedialog.askopenfilename(initialdir=self.__initialdir,
                                                      title=self.__title,
                                                      **self.__cnfs[DIALOG])
            except Exception as e:
                return
            path = None
            if not filename:
                pass
            elif isinstance(filename, str):
                path = filename
            else:
                path = ";".join(filename)
            if path:
                self.__string_var.set(path)
        else:
            try:
                filename = filedialog.askdirectory(initialdir=self.__initialdir,
                                                   title=self.__title,
                                                   **self.__cnfs[DIALOG])
            except Exception as e:
                return
            path = None
            if not filename:
                pass
            elif isinstance(filename, str):
                path = filename
            else:
                path = ";".join(filename)
            if path:
                self.__string_var.set(path)
        self.__entry.icursor("end")


if __name__ == "__main__":
    root = tk.Tk()
    pathentry_test = Pathentry(root, browse="dir",
                               extra_options={"dialog":
                                                  {"initialdir": "/home/alex",
                                                   "title": "Hello"}})
    pathentry_test.pack(fill=tk.BOTH, expand=1)
    root.mainloop()



import sys
import time
import subprocess
import os
import os.path
import pkgutil
from pyrustic.jasonix import Jasonix
from pyrustic.manager.misc.lite_test_runner import LiteTestRunner
from pyrustic.manager.misc.funcs import setup_config, wheels_assets


class BuildHandler:
    """
    Description
    -----------
    Use this command to build a distribution package
    that could be published later with Hub.
    This command will block the Pyrustic Manager.
    The distribution package is a Wheel.

    Usage
    -----
    - Description: Build
    - Command: build
    """

    def __init__(self, target, app_pkg):
        self._target = target
        self._app_pkg = app_pkg
        self._version = None
        self._pre_building_hook = None
        self._post_building_hook = None
        self._process(target, app_pkg)

    def _process(self, target, app_pkg):
        if target is None:
            print("Please link a Target first. Check 'help target'.")
            return
        if not app_pkg:
            print("Please initialize this project first. Check 'help init'.")
            return
        # version
        self._version = self._get_version()
        # get confirmation
        message = "You are going to build '{}' v{}.".format(app_pkg,
                                                        self._version)
        print(message)
        if not self._ask_for_confirmation("\nDo you want to continue ?"):
            return
        print("")
        # tests
        if not self._run_tests():
            return
        # build
        pre, post = self._get_hooks()
        self._pre_building_hook, self._post_building_hook = pre, post
        if not self._run_pre_building_hook():
            return
        print("\nBuilding...")
        if sys.executable:
            args = ["setup.py", "--quiet", "sdist", "bdist_wheel"]
            p = subprocess.Popen([sys.executable, *args], cwd=target)
            p.communicate()
            code = p.returncode
            if code == 0:
                self._gen_build_report()
                if not self._run_post_building_hook():
                    return
                print("\nSuccessfully built !")
                return
            print("\nFailed to built the project.\nReturn code: {}".format(code))

    def _run_tests(self):
        if not self._ask_for_confirmation("Do you want to run tests first ?",
                                          "n"):
            return True
        print("Running tests...")
        test_exist, test_success, test_result = self._test_runner()
        if not test_exist:
            print("There aren't Tests\n")
            return True
        if test_success:
            print("Testing passed\n")
            return True
        else:
            print("Testing failed")
            print(test_result)
            return False

    def _test_runner(self):
        test_path = os.path.join(self._target, "tests")
        test_success = True
        test_result = None
        test_exist = False
        if os.path.exists(test_path):
            test_exist = True
            test_host = LiteTestRunner(test_path, self._target)
            test_success, test_result = test_host.run()
        return test_exist, test_success, test_result

    def _ask_for_confirmation(self, message, default="y"):  #TODO: fix, EOF error (ctrl+d)
        cache = "Y/n" if default == "y" else "y/N"
        user_input = input("{} ({}): ".format(message, cache))
        if not user_input:
            user_input = default
        if user_input.lower() == "y":
            return True
        return False

    def _run_pre_building_hook(self):
        if not self._pre_building_hook:
            return True
        args = ["-m", self._pre_building_hook, self._target,
                self._app_pkg, self._version]
        p = subprocess.Popen([sys.executable, *args],
                             cwd=self._target)
        p.communicate()
        code = p.returncode
        print("")
        if code == 0:
            print("pre_building_hook.py executed with success")
            return True
        else:
            print("Failed to execute pre_building_hook.py")
            return False

    def _run_post_building_hook(self):
        if not self._post_building_hook:
            return True
        args = ["-m", self._post_building_hook, self._target,
                self._app_pkg, self._version]
        p = subprocess.Popen([sys.executable, *args],
                             cwd=self._target)
        p.communicate()
        code = p.returncode
        print("")
        if code == 0:
            print("post_building_hook.py executed with success")
            return True
        else:
            print("Failed to execute post_building_hook.py")
            return False

    def _gen_build_report(self):
        pyrustic_data_path = os.path.join(self._target,
                                          self._app_pkg,
                                          "pyrustic_data")
        try:
            os.mkdir(pyrustic_data_path)
        except Exception as e:
            pass
        res = "manager/default_json/pyrustic_data/build_report_default.json"
        default_build_report_json = pkgutil.get_data("pyrustic", res)
        build_report_json = os.path.join(pyrustic_data_path,
                                         "build_report.json")
        jasonix = Jasonix(build_report_json,
                          default=default_build_report_json)
        jasonix.data["timestamp"] = int(time.time())
        jasonix.data["target"] = self._target
        jasonix.data["app_pkg"] = self._app_pkg
        wheels_assets_list = wheels_assets(self._target)
        wheel_asset = None
        if wheels_assets_list:
            wheel_asset = wheels_assets_list[0]
            wheel_asset = os.path.join(self._target,
                                       "dist",
                                       wheel_asset)
        jasonix.data["app_version"] = self._version
        jasonix.data["wheel_asset"] = wheel_asset
        jasonix.save()

    def _get_hooks(self):
        dev_json = os.path.join(self._target,
                                self._app_pkg,
                                "pyrustic_data",
                                "dev.json")
        if not os.path.exists(dev_json):
            return None, None
        jasonix = Jasonix(dev_json)
        hooks_package = jasonix.data.get("hooking_pkg", None)
        if not hooks_package:
            return None, None
        pre_building_hook = "{}.pre_building_hook".format(hooks_package)
        post_building_hook = "{}.post_building_hook".format(hooks_package)
        return pre_building_hook, post_building_hook

    def _get_version(self):
        setup_config_dict = setup_config(self._target)
        version = None
        if setup_config_dict:
            version = setup_config_dict["version"]
        return version



import os
import os.path
import json
try:
    from pyrustic.exception import PyrusticException
except ImportError:
    class PyrusticException(Exception):
        pass


class Jasonix:
    """
    Jasonix allows you to play with JSON files like toys ! (really)
    """
    def __init__(self, target, default=None, readonly=False):
        """
        PARAMETERS:

        - target: dict or file-like object or a path to a json file. If target is a path
         and this path doesn't exist, a new file will be created or not according
        to the parameter "default

        - default: file-like object or a path or a dict.

        - readonly: bool
        """
        self._target = target
        self._default = default
        self._readonly = readonly
        #
        self._data = None
        #
        self._setup()


    # ==============================================
    #               PROPERTIES
    # ==============================================

    @property
    def data(self):
        """
        The dict-like representation of the JSON file
        """
        return self._data

    @data.setter
    def data(self, val):
        """
        The dict to push into JSON file
        """
        self._data = val

    @property
    def target(self):
        return self._target

    @property
    def default(self):
        return self._default

    # ==============================================
    #               PUBLIC METHODS
    # ==============================================

    def save(self):
        """"
        Push data into the JSON file (not the default file !) if 'readonly' is False
        """
        if self._readonly:
            PyrusticException("Attempt to save a readonly config !")
        self._json_dump(self._target, self._data)

    def reload(self):
        """
        Reload data from JSON file
        """
        self._setup()

    # ==============================================
    #               PRIVATE METHODS
    # ==============================================
    def _setup(self):
        if not self._target:
            return
        data = self._json_load(self._target,
                               ignore_exception=True)
        if data is None:
            if isinstance(self._target, str):
                with open(self._target, "w") as file:
                    pass
            if not self._default:
                raise PyrusticException("Missing target !")
            default_data = self._json_load(self._default)
            if default_data is None:
                message = ("Missing target !",
                           "And invalid default json.")
                raise PyrusticException(" ".join(message))
            self._data = default_data
            self.save()
            return
        self._data = data

    def _json_load(self, target, ignore_exception=False):
        data = None
        if isinstance(target, str):
            data = self._json_load_from_path(target,
                                             ignore_exception)
        elif isinstance(target, dict):
            data = target
        elif isinstance(target, bytes):
            target = target.decode("utf-8")
            data = json.loads(target)
        else:
            data = self._json_load_from_file(target,
                                             ignore_exception)
        return data

    def _json_load_from_path(self, path,
                             ignore_exception=False):
        data = None
        try:
            with open(path, "r") as file:
                data = json.load(file)
        except Exception as e:
            if not ignore_exception:
                raise e
        return data

    def _json_load_from_file(self, file,
                             ignore_exception=False):
        data = None
        try:
            data = json.load(file)
        except Exception as e:
            if not ignore_exception:
                raise e
        return data

    def _json_dump(self, target, data,
                   ignore_exception=False):
        if isinstance(target, str):
            self._json_dump_to_path(target, data, ignore_exception)
        else:
            self._json_dump_to_file(target, data, ignore_exception)

    def _json_dump_to_path(self, path, data,
                           ignore_exception=False):
        try:
            with open(path, "w") as file:
                json.dump(data, file, indent=4, sort_keys=True)
        except Exception as e:
            if not ignore_exception:
                raise e

    def _json_dump_to_file(self, file, data,
                           ignore_exception=False):
        try:
            json.dump(data, file, indent=4, sort_keys=True)
        except Exception as e:
            if not ignore_exception:
                raise e





class Theme:
    """
    A theme is a collection of styles and... others theme.
    """
    def __init__(self):
        self._styles = []
        self._themes = []

    @property
    def styles(self):
        """
        Get a list of styles that this theme contains
        """
        return self._styles.copy()

    @property
    def themes(self):
        """
        Get a list of theme that this theme contains
        """
        return self._themes.copy()

    def add_style(self, style, scope=None):
        """
        Add a style to this theme.
        The style is an instance of 'pyrustic.default_style._Style'.
        You don't have to directly subclass the private class '_Style'.
        Instead, subclass one of the public classes in the module 'pyrustic.default_style'.
        The scope here is an optional string.
        When you don't set the scope, the style will be applied as it.
        Example. If you add a Button style in your theme, this style will be
        applied to all buttons widgets. But you can restrict this effect to a scope.
        This scope could be by example "*Canvas*Button.", meaning all buttons
        that are living on all Canvas, are candidates for the given style.
        """
        self._styles.append((style, scope))

    def add_theme(self, theme):
        """
        Add a theme to this... theme
        """
        self._themes.append(theme)
        self._styles.extend(theme.styles)

    def target(self, master):
        """
        Set this theme to master. Master here should be the root widget of your app.
        You need to set the theme to master before installing others widgets on the master.
        """
        for style, scope in self._styles:
            widget_class = style.widget_class
            for key, val in style.__dict__.copy().items():
                if val is None:
                    continue
                pattern = None
                if scope is None:
                    pattern = "*{}.{}".format(widget_class, key)
                else:
                    pattern = "{}{}".format(scope, key)
                master.option_add(pattern, val)



import os
import os.path
import pkgutil
from pyrustic.jasonix import Jasonix
from pyrustic.manager import constant


def main():
    set_environment()


def set_environment():
    success = True
    # Make folders
    if not make_folders():
        success = False
    # Add shared data files
    if success and not add_default_shared_data_files():
        success = False
    # check success
    if success:
        return True
    else:
        return False


def make_folders():
    folders = (constant.PYRUSTIC_DATA_FOLDER,
               constant.MANAGER_SHARED_FOLDER,
               constant.MANAGER_CACHE_FOLDER,
               constant.RUNTEST_SHARED_FOLDER,
               constant.RUNTEST_CACHE_FOLDER,
               constant.SQLEDITOR_SHARED_FOLDER,
               constant.SQLEDITOR_CACHE_FOLDER,
               constant.HUB_SHARED_FOLDER,
               constant.HUB_CACHE_FOLDER)
    for path in folders:
        if not _make_folder(path):
            return False
    return True


def add_default_shared_data_files():
    resource_prefix = "manager/default_json/PyrusticData/"
    data = (("manager_default.json", constant.MANAGER_SHARED_DATA_FILE),
            ("jupitest_default.json", constant.RUNTEST_SHARED_DATA_FILE),
            ("rustiql_default.json", constant.SQLEDITOR_SHARED_DATA_FILE),
            ("hubway_default.json", constant.HUB_SHARED_DATA_FILE))
    for json_name, path in data:
        data = pkgutil.get_data("pyrustic",
                                resource_prefix + json_name)
        if os.path.exists(path):
            return True
        try:
            Jasonix(path, default=data)
        except Exception as e:
            print("Failed to initialize {}".format(json_name))
            print(e)
    return True


def _make_folder(path):
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except Exception as e:
            print("Failed to make directory '{}'".format(path))
            print(e)
            return False
    return True


if __name__ == "__main__":
    main()



import os
import os.path
import pkgutil
from pyrustic.manager.misc import funcs
from pyrustic.jasonix import Jasonix


class InitHandler:
    """
    Description
    -----------
    Use this command to init your project.
    Pyrustic Manager will install a basic
    project structure in your project.
    The PROJECT_DIR is the project's root
    directory.
    The APP_DIR is the directory of your
    source code.
    The APP_PKG is simply the name of the
    root package of your source code.

    Usage
    -----
    - Description: Init your project
    - Command: init

    Example
    -------
    Assume the linked target is:
    /home/alex/demo
    This target is also your project root
    directory. And 'demo' is your project's
    name. So let's assume that your target
    is an empty directory.
    When you issue the command 'init', this
    is what the project root will look like:
    demo # target or PROJECT_ROOT
        demo # APP_PKG or APP_DIR, source here
            __main__.py # entry point
            __init__.py
            version.py # __version__ = "0.0.1"
            view # the demo.view package
                main_view.py # module
            pyrustic_data # folder
                hubstore.json
                gui.json # configure your GUI
        tests
            __init__.py
        setup.py
        setup.cfg # edit your project config
        pyproject.toml
        MANIFEST.in # don't worry, I take care

    So when you want to add a file "my_file.txt"
    and the module "mod.py" in the package
    demo.view, you issue in the manager:
        - add demo.view my_file.txt mod.py

    """
    def __init__(self, target, app_pkg, args):
        self._target = target
        self._app_pkg = app_pkg
        self._process(args)

    def _process(self, args):
        if not self._target:
            print("You should link a Target first")
            return
        if args:
            print("Wrong usage of this command")
            return
        # ask for app_pkg
        self._set_app_pkg()
        # create package
        self._make_packages()
        # create folders
        self._make_folders()
        # add files
        self._add_files()
        # add json data files
        self._add_json_data_files()
        print("Successfully initialized !")

    def _make_packages(self):
        hooking_pkg = "{}.hooking".format(self._app_pkg)
        packages = (self._app_pkg, "tests", hooking_pkg)
        for package in packages:
            funcs.build_package(self._target, package)
        app_dir = os.path.join(self._target, self._app_pkg)
        packages = ("view", )
        for package in packages:
            funcs.build_package(app_dir, package)

    def _make_folders(self):
        folders = ("pyrustic_data",)
        for folder in folders:
            path = os.path.join(self._target, self._app_pkg, folder)
            if os.path.exists(path):
                continue
            os.mkdir(path)

    def _add_files(self):
        resource_prefix = "manager/template/"
        # add version.py
        resource = resource_prefix + "version_template.txt"
        app_dir = os.path.join(self._target, self._app_pkg)
        dest_path = os.path.join(app_dir, "version.py")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)
        # add __main__.py
        resource = resource_prefix + "main_template.txt"
        dest_path = os.path.join(app_dir, "__main__.py")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        data = data.format(app_pkg=self._app_pkg)
        self._add_file(dest_path, data)
        # add main_view.py
        resource = resource_prefix + "main_view_template.txt"
        dest_path = os.path.join(app_dir, "view", "main_view.py")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)
        # add .gitignore
        resource = resource_prefix + "gitignore_template.txt"
        dest_path = os.path.join(self._target, ".gitignore")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)
        # add LICENSE
        resource = resource_prefix + "license_template.txt"
        dest_path = os.path.join(self._target, "LICENSE")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)
        # add README.md
        resource = resource_prefix + "readme_template.txt"
        dest_path = os.path.join(self._target, "README.md")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)
        # add MANIFEST.in
        resource = resource_prefix + "manifest_template.txt"
        dest_path = os.path.join(self._target, "MANIFEST.in")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        data = data.format(app_pkg=self._app_pkg)
        self._add_file(dest_path, data)
        # add setup.py
        resource = resource_prefix + "setup_py_template.txt"
        dest_path = os.path.join(self._target, "setup.py")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)
        # add setup.cfg
        resource = resource_prefix + "setup_cfg_template.txt"
        dest_path = os.path.join(self._target, "setup.cfg")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        data = data.format(project_name=os.path.basename(self._target),
                           app_pkg=self._app_pkg)
        self._add_file(dest_path, data)
        # add pyproject.toml
        resource = resource_prefix + "pyproject_template.txt"
        dest_path = os.path.join(self._target, "pyproject.toml")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)
        # add pre_building_hook.py
        resource = resource_prefix + "pre_building_hook_template.txt"
        dest_path = os.path.join(self._target, self._app_pkg,
                                 "hooking",
                                 "pre_building_hook.py")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)
        # add post_building_hook.py
        resource = resource_prefix + "post_building_hook_template.txt"
        dest_path = os.path.join(self._target, self._app_pkg,
                                 "hooking",
                                 "post_building_hook.py")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)
        # add pre_publishing_hook.py
        resource = resource_prefix + "pre_publishing_hook_template.txt"
        dest_path = os.path.join(self._target, self._app_pkg,
                                 "hooking",
                                 "pre_publishing_hook.py")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)
        # add post_publishing_hook.py
        resource = resource_prefix + "post_publishing_hook_template.txt"
        dest_path = os.path.join(self._target, self._app_pkg,
                                 "hooking",
                                 "post_publishing_hook.py")
        data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
        self._add_file(dest_path, data)

    def _add_json_data_files(self):
        local_pyrustic_data_folder = os.path.join(self._target,
                                                  self._app_pkg,
                                                  "pyrustic_data")
        resource_prefix = "manager/default_json/pyrustic_data/"
        # add dev.json
        path = os.path.join(local_pyrustic_data_folder, "dev.json")
        default_resource = resource_prefix + "dev_default.json"
        data = pkgutil.get_data("pyrustic", default_resource)
        if not os.path.exists(path):
            with open(path, "wb") as file:
                file.write(data)
        jasonix = Jasonix(path)
        jasonix.data["hooking_pkg"] = "{}.hooking".format(self._app_pkg)
        jasonix.save()
        # add gui.json
        path = os.path.join(local_pyrustic_data_folder, "gui.json")
        default_resource = resource_prefix + "gui_default.json"
        data = pkgutil.get_data("pyrustic", default_resource)
        if not os.path.exists(path):
            with open(path, "wb") as file:
                file.write(data)
        # add publishing.json
        path = os.path.join(local_pyrustic_data_folder,
                            "publishing.json")
        default_resource = resource_prefix + "publishing_default.json"
        data = pkgutil.get_data("pyrustic", default_resource)
        if not os.path.exists(path):
            with open(path, "wb") as file:
                file.write(data)
        # add hubstore.json
        path = os.path.join(local_pyrustic_data_folder, "hubstore.json")
        default_resource = resource_prefix + "hubstore_default.json"
        data = pkgutil.get_data("pyrustic", default_resource)
        if not os.path.exists(path):
            with open(path, "wb") as file:
                file.write(data)

    def _add_file(self, path, data):
        if os.path.exists(path):
            return
        with open(path, "w") as file:
            file.write(data)

    def _set_app_pkg(self):
        if self._app_pkg is not None:
            return
        self._app_pkg = os.path.basename(self._target)



"""
Pyrustic exceptions
"""


class PyrusticException(Exception):
    pass

class PyrusticAppException(PyrusticException):
    pass


class PyrusticWidgetException(PyrusticException):
    pass


class PyrusticTableException(PyrusticWidgetException):
    pass



import threading
import collections
import queue


class Threadom:
    def __init__(self, tk, sync=False):
        """
        - tk: a tk.Tk instance or any tkinter object
        - sync: boolean
        """
        self._tk = tk
        self._sync = sync
        self._internal_count = 0
        self._queues = dict()
        self._queues_lock = threading.Lock()
        self._running_synced_target_lock = threading.Lock()
        self._waiting_synced_target_to_run = list()
        self._running_synced_target = False


    # ===========================================
    #               PUBLIC METHODS
    # ===========================================
    def q(self):
        """
        Creates a new Queue
        """
        return queue.Queue()

    def consume(self, queue, consumer=None, unpack_result=False,
                exception_handler=None, latency=10):
        """
        Loops through the queue, pick data, then run the callback 'consumer' with
        data as argument.
        Example, assume that there are these integers 3, 4 and 5 in the queue.
        3 -> consumer(3); 4 -> consumer(4); 5 -> consumer(5)

        - queue: the queue. See Threadom's method 'q'.
        - consumer: the callback that accepts one argument or more than one if unpack_result is True
        - unpack_result: if True, the result will be unpacked.
        - exception_handler: callback that accepts one argument to handle any occurred exception.
        - latency: integer. Milliseconds between each loop to consume the queue. By default: 10

        Returns the 'qid'. You will need this 'qid' to stop, pause, resume the loop or to get info.
        """
        qid = self._consume(queue,
                            consumer,
                            unpack_result,
                            exception_handler,
                            latency)
        return qid

    def pause(self, qid):
        """
        Pause the process launched by the method 'consume'.
        Put 0 to pause all processes
        """
        if qid == 0:
            with self._queues_lock:
                for key in self._queues.keys():
                    self._queues[key]["active"] = False
        elif self._is_valid_qid(qid):
            with self._queues_lock:
                self._queues[qid]["active"] = False

    def resume(self, qid):
        """
        Resume the process launched by the method 'consume'
        Put 0 to resume all processes
        """
        if qid == 0:
            with self._queues_lock:
                for key in self._queues.keys():
                    if not self._queues[key]["active"]:
                        self._queues[key]["active"] = True
                        self._loop(key)
        elif self._is_valid_qid(qid):
            with self._queues_lock:
                if not self._queues[qid]["active"]:
                    self._queues[qid]["active"] = True
                    self._loop(qid)

    def stop(self, qid):
        """
        Stop the process launched by the method 'consume'.
        Set 0 to stop them all.
        """
        if qid == 0:
            with self._queues_lock:
                self._queues = dict()
        elif self._is_valid_qid(qid):
            with self._queues_lock:
                del self._queues[qid]

    def info(self, qid=None):
        """
        Retrieve info from the process launched by the method 'consume'.
        Returns a dict:
         {"queue": queue, "active": boolean, "consumer": callback, "unpack_result": boolean,
         "exception_handler": callback, "latency": integer}
        """
        if qid is None:
            with self._queues_lock:
                return tuple([x.copy() for x in self._queues])
        if not self._is_valid_qid(qid):
            return
        with self._queues_lock:
            return self._queues[qid].copy()

    def run(self, target, target_args=None,
            target_kwargs=None,
            consumer=None,
            sync=None,
            daemon=True,
            unpack_result=False,
            upstream_exception_handler=None,
            downstream_exception_handler=None):
        """
        Runs a target in background. Return False if the target is in WAITING state (sync)
        - target: the callable to run
        - target_args: tuple, arguments to use
        - target_kwargs: dict, keyword-arguments to use
        - consumer: the callback with parameter(s) that will consume the returned value by target
        - sync: None or boolean to override the constructor's argument sync
        - unpack_result: boolean, True, to unpack the result returned by target
        - upstream_exception_handler: one parameter callback to handle the exception
            raised while running the target
        - downstream_exception_handler: one parameter callback to handle the exception
            raised while calling the consumer
        """
        sync = self._sync if sync is None else sync
        target_args = {} if target_args is None else target_args
        target_kwargs = {} if target_kwargs is None else target_kwargs
        return self._run(target, target_args,
                         target_kwargs, consumer,
                         sync, daemon, unpack_result,
                         upstream_exception_handler,
                         downstream_exception_handler)

    # ===========================================
    #               INTERNAL
    # ===========================================

    def _run(self, target, target_args,
             target_kwargs, consumer,
             sync, daemon, unpack_result,
             upstream_exception_handler,
             downstream_exception_handler):
        if sync:
            with self._running_synced_target_lock:
                if self._running_synced_target:
                    data = {"target": target,
                            "target_args": target_args,
                            "target_kwargs": target_kwargs,
                            "consumer": consumer,
                            "sync": sync, "daemon": daemon,
                            "unpack_result": unpack_result,
                            "upstream_exception_handler":
                                upstream_exception_handler,
                            "downstream_exception_handler":
                                downstream_exception_handler}
                    self._waiting_synced_target_to_run.append(data)
                    return False
                self._running_synced_target = True
        queue = self.q()
        thread = threading.Thread(target=self._runner,
                                  args=(queue, target, target_args,
                                        target_kwargs,
                                        upstream_exception_handler),
                                  daemon=daemon)
        thread.start()
        self._short_loop(queue, consumer, unpack_result,
                         upstream_exception_handler,
                         downstream_exception_handler)
        return True

    def _consume(self, queue, consumer, unpack_result,
                exception_handler, latency):
        if queue is None:
            return
        self._internal_count += 1
        qid = self._internal_count
        data = dict()
        data["queue"] = queue
        data["active"] = True
        data["consumer"] = consumer
        data["unpack_result"] = unpack_result
        data["exception_handler"] = exception_handler
        data["latency"] = latency
        self._queues[qid] = data
        self._loop(qid)
        return qid

    def _runner(self, queue, target, target_args, target_kwargs, exception_handler):
        result = None
        exception = None
        exception_occurred = False
        try:
            result = target(*target_args, **target_kwargs)
        except Exception as e:
            exception_occurred = True
            exception = e
        if not exception_occurred:
            queue.put(result)
        elif exception_occurred and exception_handler is not None:
            queue.put(None)
            queue.put(exception)
        else:
            raise exception

    def _loop(self, qid):
        if not self._is_valid_qid(qid):
            return
        data = self._queues[qid]
        if not data["active"]:
            return
        queue = data["queue"]
        consumer = data["consumer"]
        unpack_result = data["unpack_result"]
        exception_handler = data["exception_handler"]
        latency = data["latency"]
        if not queue.empty():
            result = queue.get()
            self._dispatch_result(result, consumer, unpack_result, exception_handler)
            if result is QueueTail:
                return
        next_call = lambda self=self, qid=qid: self._loop(qid)
        self._tk.after(latency, func=next_call)

    def _short_loop(self, queue, consumer,
                    unpack_result,
                    upstream_exception_handler,
                    downstream_exception_handler):
        if not queue.empty():
            result = queue.get()
            exception = None
            if not queue.empty():
                exception = queue.get()
            if exception is not None:
                self._dispatch_exception(exception, upstream_exception_handler)
            elif exception is None and consumer is not None:
                self._dispatch_result(result, consumer,
                                      unpack_result,
                                      downstream_exception_handler)
            self._run_next_synced_target()
        else:
            next_call = (lambda self=self, queue=queue, consumer=consumer,
                            unpack_result=unpack_result,
                            upstream_exception_handler=upstream_exception_handler,
                            downstream_exception_handler=downstream_exception_handler:
                                self._short_loop(queue, consumer,
                                                 unpack_result,
                                                 upstream_exception_handler,
                                                 downstream_exception_handler))
            self._tk.after(5, next_call)

    def _dispatch_result(self, result, consumer, unpack_result, exception_handler):
        if consumer is not None:
            try:
                self._run_consumer(consumer, result, unpack_result)
            except Exception as e:
                if exception_handler is not None:
                    try:
                        exception_handler(e)
                    except Exception as e:
                        raise e
                else:
                    raise e

    def _dispatch_exception(self, exception, exception_handler):
        if exception_handler is not None:
            try:
                exception_handler(exception)
            except Exception as e:
                raise e
        else:
            raise exception

    def _run_next_synced_target(self):
        self._running_synced_target = False
        if self._waiting_synced_target_to_run:
            data = self._waiting_synced_target_to_run[0]
            del self._waiting_synced_target_to_run[0]
            self._run(data["target"], data["target_args"], data["target_kwargs"],
                      data["consumer"], data["sync"], data["daemon"],
                      data["unpack_result"], data["upstream_exception_handler"],
                      data["downstream_exception_handler"])

    def _is_valid_qid(self, qid):
        with self._queues_lock:
            if qid is not None and qid in self._queues:
                return True
            return False

    def _run_consumer(self, consumer, result, unpack_result):
        if unpack_result and isinstance(result, dict):
            consumer(**result)
        elif unpack_result and isinstance(result, collections.Sequence):
            consumer(*result)
        else:
            consumer(result)


class QueueTail:
    pass



import tkinter as tk
import operator
from pyrustic.exception import PyrusticTableException
from pyrustic import widget
from pyrustic.tkmisc import merge_cnfs


# Allowed Options for columns
COLUMN_OPTIONS = ["background", "borderwidth", "cursor",
                  "disabledforeground", "exportselection", "font",
                  "foreground", "height", "highlightbackground",
                  "highlightcolor", "highlightthickness", "justify",
                  "relief", "selectbackground", "selectborderwidth",
                  "selectforeground", "setgrid", "state", "takefocus",
                  "width"]

# Components of table
VSB = "vsb"
HSB = "hsb"
CANVAS = "canvas"
FRAME_BACKGROUND = "frame_background"
FRAMES_HEADERS = "frames_headers"
LISTBOXES_COLUMNS = "listboxes_columns"
LABELS_SORTING = "labels_sorting"
LABELS_TITLES = "labels_titles"

# Layout
EQUALLY = "equally"
PROPORTIONALLY = "proportionally"

# Selection modes for rows
SINGLE = "single"
BROWSE = "browse"
MULTIPLE = "multiple"
EXTENDED = "extended"

# Orient
BOTH = "both"
VERTICAL = "vertical"
HORIZONTAL = "horizontal"

# sort data as
ASC = "[+]"
DESC = "[-]"


class Table(widget.Frame):  # TODO the select_mode MULTIPLE is buggy !
    """
    Table supports data sorting, multiple selection modes, and more...

    Example:
    ```python
    import tkinter as tk
    from pyrustic.widget.table import Table

    root = tk.Tk()
    my_titles = ("Name", "Job")
    my_data = (("Jack", "Architect"), ("Diana", "Physicist"))
    table = Table(root, titles=my_titles, data=my_data)
    table.build_pack()
    root.mainloop()
    ```

    """
    def __init__(self,
                 master=None,
                 titles=None,
                 data=None,
                 hidden_columns=None,
                 sorting=True,
                 mask=None,
                 select_mode=BROWSE,
                 layout=EQUALLY,
                 orient=BOTH,
                 cnfs=None):
        """
        PARAMETERS:

        - master: widget parent. Example: an instance of tk.Frame

        - titles: sequence of titles. Example: ("Name", "Job")

        - data: sequence of sequences. Each sub-sequence must have same size as titles.
            Example: ( ("Jack, "Architect"), ("Diana", "Physicist") )

        - hidden_columns: sequence of columns to hide.
            Example: (1, 2) will hide the column at the index 1 and 2.
            Example: (0, ) will hide only the first column

        - sorting: boolean, set to True if you want the table to be able to do sorting when user
            clicks on a column title. Default: True

        - mask: a callable that will be called at each insertion of line of data
        in the table.
            The mask must accept 2 arguments:
                - index: int, index of the row (line)
                - data: the sequence of strings to insert at this given row
            The mask must returns a new data with same length or the same old data

        - select_mode: selection modes: SINGLE, BROWSE,
         MULTIPLE and EXTENDED. Default: SINGLE.
         Selection modes are the same as described in the tk.Listbox's documentation.

        - layout: EQUALLY or PROPORTIONALLY. Default: EQUALLY

        - orient: orientation for scrollbars. BOTH or VERTICAL or HORIZONTAL

        - options: dictionary of widgets options
            The widgets keys are: BODY, VSB, HSB, CANVAS, FRAME_BACKGROUND,
            FRAMES_HEADERS, LISTBOXES_COLUMNS, LABELS_SORTING and LABELS_TITLES.
            Example: Assume that you want to set the BODY's background to black
            and the horizontal scrollbar's background to red:
                options = {"BODY": {"background": "red"},
                           "HSB": {"background": "black"}}
        """
        self.__cnfs = merge_cnfs(None, cnfs,
                                 components=("body", VSB, HSB, CANVAS,
                        FRAME_BACKGROUND, FRAMES_HEADERS, LISTBOXES_COLUMNS,
                        LABELS_SORTING, LABELS_TITLES))
        super().__init__(master=master,
                         class_="Table",
                         cnf=self.__cnfs["body"],
                         on_build=self.__on_build,
                         on_display=self.__on_display,
                         on_destroy=self.__on_destroy)
        # check if listboxes options are valid
        _verify_options(self.__cnfs[LISTBOXES_COLUMNS])
        self.__titles_cache = () if titles is None else titles
        self.__titles = []
        self.__data_cache = () if data is None else data
        self.__data = []
        self.__hidden_columns = () if hidden_columns is None else hidden_columns
        self.__sorting = sorting
        self.__mask = mask
        self.__select_mode = select_mode
        self.__layout = layout
        self.__orient = orient
        self.__canvas_options = None
        self.__frame_background_options = None
        self.__frames_headers_options = None
        self.__labels_sorting_options = None
        self.__labels_titles_options = None
        self.__listboxes_columns_options = None
        # misc
        self.__components = {}
        self.__cache = None
        self.__current_sorting = None
        self.__current_column_index = None
        self.__current_row_index = None
        self.__selection_garbage = None
        self.__selection = None
        self.__header_clicked_handlers = []
        self.__header_event_handlers = {}
        self.__row_selected_handlers = []
        self.__row_event_handlers = {}
        self.__default_listbox_background = None
        self.__default_listbox_foreground = None
        self.__default_listbox_selectbackground = None
        self.__default_listbox_selectforeground = None
        # cache for sorting's labels, header's labels and listboxes
        self.__header_frames_cache = []
        self.__labels_sorting_cache = []
        self.__labels_titles_cache = []
        self.__listboxes_cache = []
        # string vars cache
        self.__labels_sorting_stringvars_cache = []
        self.__labels_titles_stringvars_cache = []
        # components
        self.__canvas = None
        self.__background = None
        self.__background_id = None
        self.__vsb = None
        self.__hsb = None
        self.__hsb_under_mouse = False
        # Sorry but the select_mode MULTIPLE is buggy
        if select_mode == MULTIPLE:
            raise PyrusticTableException("Sorry but the selection mode MULTIPLE is buggy")
        self.__view = self.build()

    # ==============================================
    #                   PROPERTIES
    # ==============================================
    @property
    def titles(self):
        return self.__titles

    @titles.setter
    def titles(self, titles):
        """
        Titles are a sequence of strings. This property overwrite the existing titles.
        """
        self.__reset_titles(titles)

    @property
    def data(self):
        return self.__data

    @data.setter
    def data(self, data):
        """
        Data is a sequence of sequences of strings.
        This property overwrite the existing data.
        Example:
            Assume that the titles are: ("Name", "Age")
            Data: ( ("Jack", 56), ("Jane", 47) )
        """
        self.__reset_data(data)

    @property
    def hidden_columns(self):
        return self.__hidden_columns

    @hidden_columns.setter
    def hidden_columns(self, val):
        """
        val: sequence of indexes to hide.
        Warning, even if you want to hide just one index,
        you should put this index into a tuple or list.
        Example: hide the column of index 1: (1, ) or [1]
        """
        self.__hidden_columns = val

    @property
    def mask(self):
        return self.__mask

    @mask.setter
    def mask(self, val):
        """
        val: a callable that will be called at each insertion of line of data
        in the table.
            The mask must accept 2 arguments:
                - index: int, index of the row (line)
                - data: the sequence of strings to insert at this given row
            The mask must returns a new data with same length or the same old data
        """
        self.__mask = val

    @property
    def select_mode(self):
        return self.__select_mode

    @property
    def layout(self):
        return self.__layout

    @property
    def orient(self):
        return self.__orient

    @property
    def table_size(self):
        """
        returns the length of columns and rows: (rows, cols)
        Example:
            Assume that the table has 3 columns and 10 rows,
            this property will return (10, 3)
        """
        return len(self.__data), len(self.__titles)

    @property
    def components(self):
        """
        Get the components (widgets instances) used to build this dialog.

        This property returns a dict. The keys are:
            BODY, VSB, HSB, CANVAS, FRAME_BACKGROUND,
            FRAMES_HEADERS, LISTBOXES_COLUMNS, LABELS_SORTING and LABELS_TITLES
        Warning: FRAMES_HEADERS, LABELS_TITLES, LABELS_SORTING
         and LISTBOXES_COLUMNS are sequences of widgets by index
        """
        return self.__components

    @property
    def selection(self):
        """
        Return a sequence of the current selection.
        selection = ( item_1, item_2, ...)
        item_i = {"index": int, "data": data}
        data = a sequence of string representing the row at the index.
        """
        if not self.__selection:
            return ()
        data = []
        for i in self.__selection:
            cache = {"index": i, "data": self.__data[i]}
            data.append(cache)
        return tuple(data)


    # ==============================================
    #                 PUBLIC METHODS
    # ==============================================
    def fill(self, titles=None, data=None):
        """
        This will overwrite the titles and/or data with the new given titles or data
        """
        if titles:
            self.titles = titles
        if data:
            self.data = data

    def insert(self, index, data):
        """
        Insert into the table this data at this index.
        Index is an integer or the string "end" (meaning, put the data at the end of table).
        This method doesn't wipe the previous data stored at this index but instead,
        pull that data down.

        data is a sequence of sequences of strings.

        Example:
        Assume you want to push the new line ("Matrix", "Cameraman") at index 0.
            insert(0, [("Matrix", "Cameraman")])
        Assume you want to push ("Matrix", "Cameraman") and ("Diana", "Seller")
        at index "end".
            insert("end", [("Matrix", "Cameraman"), ("Diana", "Seller")])
        """
        if self.__check_data_row_size(data):
            self.__insert(index, data)
            self.__adjust_selection_after_insertion(index, len(data))

    def get(self, index_start, index_end=None):
        """
        Returns a line if you don't give a 'index_end'.
        Returns a sequence of lines if you give a 'index_end'.
        """
        if not index_end:
            return self.__data[index_start]
        else:
            result = []
            index_end_range = len(self.__data) if index_end == "end" else index_end + 1
            for i in range(index_start, index_end_range):
                result.append(self.__data[i])
            return result

    def delete(self, index_start, index_end=None):
        """
        Deletes lines (rows) from the table
        """
        if not index_end:
            del self.__data[index_start]
        else:
            index_fin_range = len(self.__data) if index_end == "end" else index_end + 1
            for _ in range(index_start, index_fin_range):
                del self.__data[index_start]
        for listbox in self.__listboxes_cache:
            listbox.delete(index_start, index_end)
        self.__selection_garbage = None
        self.__selection = None

    def clear(self):
        """
        Clear the table
        """
        for listbox in self.__listboxes_cache:
            listbox.delete(0, "end")
        self.__data = []

    def handle_row_selected(self, handler):
        """
        This callback will be called at the event 'row selection':
            handler(table, row_data, row_index, column_index)
        """
        self.__row_selected_handlers.append(handler)

    def handle_row_event(self, sequence, handler):
        """
        This callback will be called at a specific row event (sequence = string):
            handler(table, row_data, row_index, column_index)
        """
        for i, listbox in enumerate(self.__listboxes_cache):
            listbox.bind(sequence,
                         lambda event,
                                self=self,
                                handler=handler,
                                i=i: self.__build_handler_for_row_event(event, handler, i), "+")

    def see(self, index="end"):
        """
        The table will scroll to the given index
        """
        if self.__listboxes_cache:
            self.__listboxes_cache[0].see(index)

    def config_column(self, index=None, **options):
        """
        Configure column. If index is None, all columns will be configured
        """
        if index is None:
            for x in self.__listboxes_cache:
                x.config(**options)
                x.config(highlightcolor=x.cget("highlightbackground"))
        else:
            self.__listboxes_cache[index].config(**options)
            self.__listboxes_cache[index].config(highlightcolor=self.__listboxes_cache[index].cget("highlightbackground"))

    def cget_column(self, index=None, option="background"):
        """
        If index is None, returns a sequence of options of listboxes (columns).
        Else returns the options of the column at the given index
        """
        if index is not None and option:
            return self.__listboxes_cache[index].cget(option)
        if option:
            return [x.cget(option) for x in self.__listboxes_cache]

    # ==============================================
    #                 PRIVATE METHODS
    # ==============================================
    def __on_build(self):
        # reset titles and data
        self.__reset_titles(self.__titles_cache)
        self.__reset_data(self.__data_cache)

    def __on_display(self):
        pass

    def __on_destroy(self):
        pass

    def __build_table(self):
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.__canvas = tk.Canvas(self, name=CANVAS,
                                  cnf=self.__cnfs[CANVAS])
        self.__components[CANVAS] = self.__canvas
        self.__canvas.grid(row=0, column=0, sticky="nswe")
        self.__background = tk.Frame(self.__canvas, self.__frame_background_options)
        self.__components[FRAME_BACKGROUND] = self.__background
        self.__background_id = self.__canvas.create_window(0, 0,
                                                           window=self.__background,
                                                           anchor="nw")
        self.__background.bind("<Configure>", self.__on_configure_background, "+")
        self.__build_header_and_columns()
        self.__extract_listboxes_color()
        self.__set_scrollbars()

    def __build_header_and_columns(self):
        self.__components[FRAMES_HEADERS] = []
        self.__components[LABELS_SORTING] = []
        self.__components[LABELS_TITLES] = []
        self.__components[LISTBOXES_COLUMNS] = []
        ignored_i = 0
        for i, title in enumerate(self.__titles):
            if i in self.__hidden_columns:
                ignored_i += 1
                continue
            i -= ignored_i
            # Configure Background Grid
            if self.__layout == EQUALLY:
                self.__background.columnconfigure(i, weight=1, uniform=1)
            # Build Header
            # - install header frame
            frame_header = tk.Frame(self.__background,
                                    cnf=self.__cnfs[FRAMES_HEADERS])
            self.__components[FRAMES_HEADERS].append(frame_header)
            frame_header.grid(row=0, column=i, sticky="nswe")
            frame_header.columnconfigure(1, weight=1)
            self.__header_frames_cache.append(frame_header)
            # - install sorting label
            label_sorting_stringvar = tk.StringVar()
            self.__labels_sorting_stringvars_cache.append(label_sorting_stringvar)
            label_sorting = tk.Label(frame_header,
                                     textvariable=label_sorting_stringvar,
                                     cnf=self.__cnfs[LABELS_SORTING])
            self.__components[LABELS_SORTING].append(label_sorting)
            label_sorting.grid(row=0, column=0)
            label_sorting.bind("<Button-1>",
                               lambda event, i=i: self.__on_header_clicked(event, i), "+")
            label_sorting.grid_remove()
            self.__labels_sorting_cache.append(label_sorting)
            # - install title label
            label_title_stringvar = tk.StringVar()
            label_title_stringvar.set(title)
            self.__labels_titles_stringvars_cache.append(label_title_stringvar)
            label_title = tk.Label(frame_header,
                                   textvariable=label_title_stringvar,
                                   cnf=self.__cnfs[LABELS_TITLES])
            self.__components[LABELS_TITLES].append(label_title)
            label_title.grid(row=0, column=1, sticky="nswe")
            label_title.bind("<Button-1>",
                             lambda event, i=i: self.__on_header_clicked(event, i), "+")
            self.__labels_titles_cache.append(label_title)
            # Build Columns
            listbox = tk.Listbox(self.__background,
                                 activestyle="none",
                                 selectmode=BROWSE if self.__select_mode == MULTIPLE
                                 else self.__select_mode,
                                 cnf=self.__cnfs[LISTBOXES_COLUMNS], takefocus=0)
            self.__components[LISTBOXES_COLUMNS].append(listbox)
            listbox.config(highlightcolor=listbox.cget("highlightbackground"))
            listbox.grid(row=1, column=i, sticky="nswe")
            listbox.bind('<<ListboxSelect>>',
                         lambda event, i=i: self.__on_row_selected(event, i), "+")
            self.__listboxes_cache.append(listbox)

    def __extract_listboxes_color(self):
        self.__default_listbox_background = self.__listboxes_cache[0].cget("background")
        self.__default_listbox_foreground = self.__listboxes_cache[0].cget("foreground")
        self.__default_listbox_selectbackground = self.__listboxes_cache[0].cget("selectbackground")
        self.__default_listbox_selectforeground = self.__listboxes_cache[0].cget("selectforeground")

    def __set_scrollbars(self):
        if self.__orient in (BOTH, HORIZONTAL, "x", "h"):
            self.__hsb = tk.Scrollbar(self, name=HSB,
                                      orient="horizontal",
                                      command=self.__canvas.xview)
            self.__components[HSB] = self.__hsb
            self.__hsb.grid(row=1, column=0, columnspan=2, sticky="we")
            self.__hsb.bind("<Button-4>", self.__on_mouse_wheel, "+")
            self.__hsb.bind("<Button-5>", self.__on_mouse_wheel, "+")
            self.__canvas.config(xscrollcommand=self.__hsb.set)
        if self.__orient in (BOTH, VERTICAL, "y", "v"):
            self.__vsb = tk.Scrollbar(self, name=VSB,
                                      orient="vertical",
                                      command=self.__scroll_listboxes_sync)
            self.__components[VSB] = self.__vsb
            self.__vsb.grid(row=0, column=1, sticky="ns")
        for listbox in self.__listboxes_cache:
            listbox.config(yscrollcommand=self.__scroll_listboxes_and_scrollbar_sync)

    def __on_mouse_wheel(self, event):
        # NB: event.num 5 to go down and event.num 4 to go up
        # NB: event.delta if negative, goes down, and event.delta if positive goes up
        scroll = 1 if event.num == 5 or event.delta < 0 else -1
        # NB: scroll 1 to go down and scroll -1 to go up
        if self.__orient in ("both", "horizontal", "h", "x",
                            "vertical", "v", "y"):
            self.__canvas.xview_scroll(scroll, "units")

    def __scroll_listboxes_sync(self, *args):
        for listbox in self.__listboxes_cache:
            listbox.yview(*args)

    def __scroll_listboxes_and_scrollbar_sync(self, *args):
        for listbox in self.__listboxes_cache:
            listbox.yview_moveto(args[0])
        if self.__vsb:
            self.__vsb.set(*args)

    def __on_header_clicked(self, event, i):
        if self.__titles:
            self.__update_sorting(i)
        for handler in self.__header_clicked_handlers:
            handler(self, self.__titles, i)

    def __on_row_selected(self, event, column_index):
        selection = event.widget.curselection()
        if not selection:
            return
        self.__current_row_index = selection[0]
        self.__current_column_index = column_index
        if self.__selection is not None and self.__select_mode == MULTIPLE:
            selection = self.__fix_selection_in_multiple_selectmode(selection)
        self.__manage_selection_garbage(selection)
        self.__selection = selection
        self.__sync_selection()
        # Notify handlers
        for handler in self.__row_selected_handlers:
            if self.__select_mode == SINGLE or self.__select_mode == BROWSE:
                handler(self, self.__data[self.__selection[0]], self.__selection[0], column_index)
            elif self.__select_mode == MULTIPLE or self.__select_mode == EXTENDED:
                handler(self, [self.__data[row] for row in self.__selection],
                        self.__selection, column_index)

    def __build_handler_for_row_event(self, event, handler, i):
        row_index = event.widget.nearest(event.y)
        handler(self, self.__data[row_index], row_index, i)

    def __fix_selection_in_multiple_selectmode(self, selection):
        fixed_selection = list(selection)
        if self.__current_row_index in fixed_selection:
            fixed_selection.remove(self.__current_row_index)
        else:
            fixed_selection.append(self.__current_row_index)
        return tuple(fixed_selection)

    def __manage_selection_garbage(self, selection):
        if self.__selection is None:
            return
        if self.__select_mode in (SINGLE, BROWSE):
            if self.__selection[0] != selection[0]:
                self.__selection_garbage = self.__selection
            else:
                self.__selection_garbage = None
        elif self.__select_mode == MULTIPLE:
            if len(self.__selection) > len(selection):
                self.__selection_garbage = tuple([i for i in self.__selection if i not in selection])
            else:
                self.__selection_garbage = None
        elif self.__select_mode == EXTENDED:
            self.__selection_garbage = tuple([i for i in self.__selection if i not in selection])

    def __sync_selection(self):
        if self.__select_mode == SINGLE or self.__select_mode == BROWSE:
            if self.__selection is not None:
                self.__sync_selection_for_single_or_browse_mode()
        elif self.__select_mode == MULTIPLE or self.__select_mode == EXTENDED:
            self.__sync_selection_for_multiple_or_extended_mode()

    def __sync_selection_for_single_or_browse_mode(self):
        for i, listbox in enumerate(self.__listboxes_cache):
            if self.__selection_garbage:
                listbox.itemconfig(self.__selection_garbage[0],
                                   background=self.__default_listbox_background,
                                   foreground=self.__default_listbox_foreground)
            listbox.itemconfig(self.__selection[0], background=self.__default_listbox_selectbackground,
                               foreground=self.__default_listbox_selectforeground)

    def __sync_selection_for_multiple_or_extended_mode(self):
        for i, listbox in enumerate(self.__listboxes_cache):
            if self.__selection_garbage:
                for x in self.__selection_garbage:
                    listbox.itemconfig(x, background=self.__default_listbox_background,
                                       foreground=self.__default_listbox_foreground)
            for x in self.__selection:
                listbox.itemconfig(x, background=self.__default_listbox_selectbackground,
                                   foreground=self.__default_listbox_selectforeground)
        if self.__select_mode == MULTIPLE:
            if self.__current_row_index in self.__selection:
                selectbackground = self.__default_listbox_selectbackground
                selectforeground = self.__default_listbox_selectforeground
            else:
                selectbackground = self.__default_listbox_background
                selectforeground = self.__default_listbox_foreground
            self.__listboxes_cache[self.__current_column_index].config(
                selectbackground=selectbackground,
                selectforeground=selectforeground)

    def __adjust_selection_after_insertion(self, index, len_data):
        if not self.__selection:
            return
        if index == "end":
            return
        cache = []
        for x in self.__selection:
            if x >= index:
                x += len_data
            cache.append(x)
        self.__selection = tuple(cache)

    def __update_sorting(self, i):
        if not self.__sorting:
            return
        for index, label in enumerate(self.__labels_sorting_cache):
            if index == i:
                continue
            label.grid_remove()
        for index, stringvar in enumerate(self.__labels_sorting_stringvars_cache):
            if index == i:
                continue
            stringvar.set("")
        sorting_label = self.__labels_sorting_stringvars_cache[i]
        if sorting_label.get() == "":
            self.__labels_sorting_stringvars_cache[i].set(ASC)
            self.__labels_sorting_cache[i].grid()
        elif sorting_label.get() == ASC:
            self.__labels_sorting_stringvars_cache[i].set(DESC)
            self.__labels_sorting_cache[i].grid()
        elif sorting_label.get() == DESC:
            self.__labels_sorting_stringvars_cache[i].set("")
            self.__labels_sorting_cache[i].grid_remove()
        self.__sort_data(i)

    def __sort_data(self, i):
        sorting_label_stringvar = self.__labels_sorting_stringvars_cache[i]
        if sorting_label_stringvar.get() == ASC:
            self.__cache = self.__data
            data_sorted = self.__table_sorter(self.__cache, i, len(self.__titles))
        elif sorting_label_stringvar.get() == DESC:
            data_sorted = self.__table_sorter(self.__cache, i, len(self.__titles),
                                              sorting="desc")
        elif sorting_label_stringvar.get() == "":
            data_sorted = self.__cache
            self.__cache = None
        self.__reset_data(data_sorted)

    def __table_sorter(self, data, index, count_columns, sorting=ASC):
        reverse = False if sorting == ASC else True
        data_sorted = sorted(data,
                             key=operator.itemgetter(*range(index, count_columns)),
                             reverse=reverse)
        return data_sorted

    def __reset_titles(self, titles):
        if not titles:
            return
        if not self.__titles:
            self.__titles = titles
            self.__build_table()
        elif self.__titles and len(titles) == len(self.__titles):
            for i, title in enumerate(titles):
                self.__labels_titles_stringvars_cache[i].set(title)
        else:
            raise PyrusticTableException("Incorrect length of titles")

    def __reset_data(self, data):
        if not data:
            return
        if not self.__titles:
            raise PyrusticTableException("Please submit titles first !")
        # check data
        self.__check_data_row_size(data)
        # clean listboxes
        self.clear()
        self.__insert(0, data)

    def __insert(self, index, elements):
        # update self._data
        pos = index
        if index == "end":
            pos = len(self.__data)
        self.__data[pos:pos] = elements
        if self.__mask:
            elements = [self.__mask(i, element) for i, element in enumerate(elements)]
            self.__check_data_row_size(elements)
        ignored_i = 0
        for i, col in enumerate(zip(*elements)):
            if i in self.__hidden_columns:
                ignored_i += 1
                continue
            i -= ignored_i
            column_data = [(c if c is not None else "") for c in col]
            self.__listboxes_cache[i].insert(index, *column_data)

    def __check_data_row_size(self, data):
        regular_size = len(self.__titles) - len(self.__hidden_columns)
        for row in data:
            if (len(row) - len(self.__hidden_columns)) != regular_size:
                raise PyrusticTableException("Invalid data size")
        return True

    def __on_configure_background(self, event):
        self.__canvas.config(height=self.__background.winfo_height())
        self.__canvas.config(width=self.__background.winfo_width())
        self.__canvas.config(scrollregion=self.__canvas.bbox("all"))


def _verify_options(column_options):
    for option in column_options.keys():
        if not _check_option(option):
            del column_options[option]


def _check_option(option):
    if option in COLUMN_OPTIONS:
        return True
    raise PyrusticTableException("The column option -" + option
                                 + " doesn't exist. These are legal options for column: "
                                 + str(COLUMN_OPTIONS))


if __name__ == "__main__":
    # data
    titles = ["Name", "Age"]
    data = [("Jackson", 22), ("Pollock", 57), ("John", 24), ("Poly", 79),
            ("Joyce", 21), ("Johnny", 38), ("Jack", 29), ("Bam", 69),
            ("Joshua", 23), ("Matthew", 79), ("Mateo", 42), ("Willy", 39)]
    # app
    root = tk.Tk()
    table = Table(root, titles=titles, data=data)
    table.pack()
    table.handle_row_selected(lambda *args: print(args))
    root.mainloop()



import tkinter as tk
from pyrustic import widget
from pyrustic.tkmisc import merge_cnfs
from pyrustic.view import View


# Components
CANVAS = "canvas"
BOX = "box"
HSB = "hsb"
VSB = "vsb"

# Orient
BOTH = "both"
VERTICAL = "vertical"
HORIZONTAL = "horizontal"


class Scrollbox(widget.Frame):
    """
    Scrollbox is a scrollable surface. You just need to use its property "box" as
    your layout's parent.

    Example:

        import tkinter as tk
        from pyrustic.widget.scrollbox import Scrollbox

        root = tk.Tk()
        scrollbox = Scrollbox(root)
        scrollbox.build_pack()
        # Pack 50 Label on the box
        for i in range(50):
            label = tk.Label(scrollbox.box, text="Label {}".format(i))
            label.pack(anchor=tk.W)
        root.mainloop()

    """
    def __init__(self,
                 master=None,
                 orient=VERTICAL,
                 box_sticky="nswe",
                 resizable_box=True,
                 cnfs=None):
        """
        - master: widget parent. Example: an instance of tk.Frame

        - orient: could be one of: VERTICAL, HORIZONTAL, BOTH

        - options: dictionary of widgets options
            The widgets keys are: BODY, CANVAS, BOX, HSB, VSB
            Example: Assume that you want to set the CANVAS background to red
                options = {CANVAS: {"background": "red"}}
        """
        self.__cnfs = merge_cnfs(None, cnfs,
                                 components=("body", CANVAS, BOX, HSB, VSB))
        super().__init__(master=master,
                         class_="Scrollbox",
                         cnf=self.__cnfs["body"],
                         on_build=self.__on_build,
                         on_display=self.__on_display,
                         on_destroy=self.__on_destroy)
        self.__orient = orient
        self.__box_sticky = box_sticky
        self.__resizable_box = resizable_box
        self.__canvas_options = None
        self.__canvas = None
        self.__box = None
        self.__box_id = None
        self.__vsb = None
        self.__hsb = None
        self.__hsb_under_mouse = False
        self.__is_scrollable = False
        self.__components = {}
        # build
        self.__view = self.build()
    # ==============================================
    #                   PROPERTIES
    # ==============================================


    @property
    def box(self):
        return self.__box

    @property
    def orient(self):
        return self.__orient

    @property
    def components(self):
        """
        Get the components (widgets instances) used to build this scrollbox.

        This property returns a dict. The keys are:
            BODY, CANVAS, BOX, HSB, VSB

        Warning: check the presence of key before usage. Example,
        the widget linked to the HSB key may be missing because
        only VSB is used
        """
        return self.__components

    # ==============================================
    #                 PUBLIC METHODS
    # ==============================================

    def xview_moveto(self, fraction):
        """
        Calls canvas's method 'xview_moveto'
        Set:
            - 0: to scroll to left
            - 1: to scroll to right
        """
        if self.__canvas:
            self.update_idletasks()
            self.__canvas.xview_moveto(fraction)

    def yview_moveto(self, fraction):
        """
        Calls canvas's method 'yview_moveto'
        Set:
            - 0: to scroll to top
            - 1: to scroll to bottom
        """
        if self.__canvas:
            self.update_idletasks()
            self.__canvas.yview_moveto(fraction)

    def box_config(self, **options):
        """
        As the BOX is an item compared to CANVAS, some
        the options concerning the BOX can be edited only via
        CANVAS "itemconfig" method.
        Use this method to edit these options.
        itemconfig options are: anchor, state, height, width.

        Warning: these options are not the same as the arguments
         of BOX's own constructor !
        """
        if self.__box:
            self.__canvas.itemconfig(self.__box_id, cnf=options)

    def clear(self):
        """
        Clears the Scrollbox.
        This method doesn't destruct this object but BOX's children
        """
        if self.__box:
            for x in self.__box.winfo_children():
                x.destroy()


    # ==============================================
    #                 PRIVATE METHODS
    # ==============================================

    def __on_build(self):
        self.bind("<Enter>", self.__on_enter_body, "+")
        self.bind("<Leave>", self.__on_leave_body, "+")
        self.bind("<Unmap>", self.__on_unmap_body, "+")
        self.bind("<Destroy>", self.__on_destroy_body, "+")
        self.bind_all("<MouseWheel>", self.__on_mouse_wheel, "+")
        self.bind_all("<Button-4>", self.__on_mouse_wheel, "+")
        self.bind_all("<Button-5>", self.__on_mouse_wheel, "+")
        self.columnconfigure(0, weight=1, uniform=1)
        self.rowconfigure(0, weight=1, uniform=1)
        self.winfo_toplevel().bind("<Configure>",
                                   self.__on_configure_box_canvas, "+")
        # canvas
        self.__canvas = tk.Canvas(self,
                                  name=CANVAS,
                                  width=0,
                                  height=0,
                                  cnf=self.__cnfs[CANVAS])
        self.__components[CANVAS] = self.__canvas
        self.__canvas.grid(row=0, column=0, sticky=self.__box_sticky)
        # box
        self.__box = tk.Frame(self.__canvas,
                              name=BOX,
                              cnf=self.__cnfs[BOX])
        self.__components[BOX] = self.__box
        self.__box_id = self.__canvas.create_window(0, 0, window=self.__box, anchor="nw")
        self.__box.bind("<Configure>", self.__on_configure_box_canvas, "+")
        # scrollbar
        self.__set_scrollbars()

    def __on_display(self):
        pass

    def __on_destroy(self):
        self.__unbind_funcs()

    def __on_mouse_wheel(self, event):
        if not self.__orient or not self.__is_scrollable:
            return
        # scroll down   (value: 1)   ->  event.num = 5   or  event.delta < 0
        # scroll up     (value: -1)  ->  event.num = 4   or  event.delta >= 0
        scroll = 1 if event.num == 5 or event.delta < 0 else -1
        if self.__orient in ("horizontal", "x", "h"):
            self.__canvas.xview_scroll(scroll, "units")
        elif self.__orient in ("both", "vertical", "y", "v"):
            if self.__hsb_under_mouse:
                self.__canvas.xview_scroll(scroll, "units")
            else:
                self.__canvas.yview_scroll(scroll, "units")

    def __set_scrollbars(self):
        if self.__orient in ("both", "horizontal", "h", "x"):
            self.__hsb = tk.Scrollbar(self, orient="horizontal",
                                      name=HSB,
                                      command=self.__canvas.xview,
                                      cnf=self.__cnfs[HSB])
            self.__components[HSB] = self.__hsb
            self.__hsb.grid(row=1, column=0, columnspan=2, sticky="swe")
            self.__canvas.config(xscrollcommand=self.__hsb.set)
            self.__bind_enter_leave_to_hsb()
        if self.__orient in ("both", "vertical", "v", "y"):
            self.__vsb = tk.Scrollbar(self, orient="vertical",
                                      name=VSB,
                                      command=self.__canvas.yview,
                                      cnf=self.__cnfs[VSB])
            self.__components[VSB] = self.__vsb
            self.__vsb.grid(row=0, column=1, sticky=self.__box_sticky)
            self.__canvas.config(yscrollcommand=self.__vsb.set)

    def __bind_enter_leave_to_hsb(self):
        def enter_hsb(event):
            self.__hsb_under_mouse = True
        def leave_hsb(event):
            self.__hsb_under_mouse = False
        self.__hsb.bind('<Enter>', enter_hsb, "+")
        self.__hsb.bind('<Leave>', leave_hsb, "+")

    def __on_configure_box_canvas(self, event):
        if self.__box.winfo_exists():
            if self.__orient in ("horizontal", "h", "x"):
                if self.__resizable_box:
                    self.__canvas.itemconfig(self.__box_id,
                                             height=self.__canvas.winfo_height())
                else:
                    self.__canvas.config(height=self.__box.winfo_height())
            elif self.__orient in ("vertical", "v", "y"):
                if self.__resizable_box:
                    self.__canvas.itemconfig(self.__box_id,
                                             width=self.__canvas.winfo_width())
                else:
                    self.__canvas.config(width=self.__box.winfo_width())
            self.__canvas.config(scrollregion=self.__canvas.bbox("all"))

    def __on_enter_body(self, event):
        self.__is_scrollable = True

    def __on_leave_body(self, event):
        self.__is_scrollable = False

    def __on_unmap_body(self, event):
        self.__is_scrollable = False

    def __on_destroy_body(self, event):
        self.__is_scrollable = False

    def __unbind_funcs(self):
        try:
            for val in ("<Enter>", "<Leave>",
                        "<Unmap>", "<Destroy>",
                        "<MouseWheel>", "<Button-4>",
                        "<Button-5>", "<Configure>"):
                self.unbind(val)
        except Exception as e:
            pass


class _ScrollboxTest(View):

    def __init__(self, root):
        super().__init__()
        self._root = root
        self._body = None

    def _on_build(self):
        self._body = tk.Frame(self._root)
        # Pane 1
        pane_1 = tk.Frame(self._root)
        pane_1.pack(side=tk.LEFT, padx=10,
                    pady=10, expand=1, fill=tk.BOTH)
        # Scrollbox 1
        scrollbox_1 = Scrollbox(pane_1, orient=VERTICAL)
        scrollbox_1.pack(pady=5, expand=1, fill=tk.BOTH)
        # Button 1
        command = (lambda self=self, box=scrollbox_1.box, side=tk.TOP:
                   self._on_click_add(box, side))
        button_1 = tk.Button(pane_1, text="Add",
                             command=command)
        button_1.pack(side=tk.BOTTOM)
        # Pane 2
        pane_2 = tk.Frame(self._root)
        pane_2.pack(side=tk.LEFT, padx=10,
                    pady=10, expand=1, fill=tk.BOTH)
        # Scrollbox 2
        scrollbox_2 = Scrollbox(pane_2, orient=HORIZONTAL)
        scrollbox_2.pack(pady=5, expand=1, fill=tk.BOTH)
        # Button 2
        command = (lambda self=self, box=scrollbox_2.box, side=tk.LEFT:
                   self._on_click_add(box, side))
        button_2 = tk.Button(pane_2, text="Add",
                             command=command)
        button_2.pack(side=tk.BOTTOM)

    def _on_display(self):
        pass

    def _on_destroy(self):
        pass

    def _on_click_add(self, frame, side=tk.TOP):
        label = tk.Label(frame, text="Hello Friend")
        label.pack(side=side)


if __name__ == "__main__":
    root = tk.Tk()
    scrollbox_test = _ScrollboxTest(root)
    scrollbox_test.build_pack(fill=tk.BOTH, expand=1)
    root.mainloop()



from pyrustic import pymisc
from pyrustic.manager.main import main


__all__ = ["command"]


def command(line=None, target=None):
    """
    Param:
        - line is a string
        - target is a string
    """
    print("EXECUTION")
    args = None
    if line is not None:
        args = pymisc.parse_cmd(line)
    main(argv=args, target=target)



import sys
import subprocess
import os.path


class RunHandler:
    """
    Description
    -----------
    Use this command to run a module located in the Target.
    Only dotted name of a module is allowed, so please ignore
    the extension ".py".
    Running a module blocks Pyrustic Manager.

    Usage
    -----
    - Description: Run a module
    - Command: run <the.module.name>

    - Description: Run the Target
    - Command: run
    - Note: The Manager will implicitly execute __main__.py

    - Description: Run a module with some arguments
    - Command: run <the.module.name> <argument_1> <argument_2>

    Example
    -------
    - Description: Run the module
    - Preliminary: Assume that "my_view.py" is in the "view" package
    - Command: run view.my_view

    - Description: Run the module with arguments
    - Preliminary: Assume that "my_view.py" is in the "view" package
    - Command: run view.my_view argument_1 "argument 2"

    Note: Please use simple or double quotes as delimiters if a string
    contains space
    """
    def __init__(self, target, app_pkg, args):
        self._target = target
        self._app_pkg = app_pkg
        self._process(target, app_pkg, args)

    def _process(self, target, app_pkg, args):
        name = None
        if len(args) == 0:
            if not target:
                self._print_catalog("missing_target")
                return
            elif not app_pkg:
                self._print_catalog("missing_app_pkg")
                return
            source_dir = os.path.join(target, app_pkg)
            if os.path.exists(os.path.join(source_dir, "__main__.py")):
                args = ["-m", app_pkg]
                name = "__main__"
            else:
                print("Missing entry point")
        else:
            args = ["-m", *args]
            name = " ".join(args)
        if not args:
            return
        if sys.executable:
            self._print_catalog("running", module=name)
            p = subprocess.Popen([sys.executable, *args], cwd=target)
            p.communicate()

    def _print_catalog(self, item, **kwargs):
        message = ""
        if item == "missing_target":
            message = "Please link a Target first. Check 'help target'."
        elif item == "running":
            message = "Running '{}' ...".format(kwargs["module"])
        elif item == "missing_app_pkg":
            message = "Please init the project first. Check 'help init'."
        print(message)



from pyrustic import dist


class VersionHandler:
    """
    Description
    -----------
    Use this command to check the version of Pyrustic.

    Usage
    -----
    - Description: Check information
    - Command: version
    """
    def __init__(self):
        self._process()

    def _process(self):
        print("Pyrustic Manager")
        print("Version: {}".format(dist("pyrustic")["version"]))



import tkinter as tk
from pyrustic import widget
from pyrustic.view import View, CustomView
from pyrustic.tkmisc import merge_cnfs
from pyrustic import tkmisc


# Components
LABEL_HEADER = "label_header"
LABEL_MESSAGE = "label_message"


class Toast(widget.Toplevel):
    """
    A toast is a dialog box with or without decoration
    that is displayed for a given duration.

    Any "click" action on the Toast's body will close it.

    Example:
        import tkinter as tk
        from pyrustic.widget.toast import Toast

        root = tk.Tk()
        toast = Toast(root, header="My Header", message="My Message")
        toast.build()
        root.mainloop()

    """

    def __init__(self,
                 master=None,
                 title=None,
                 header=None,
                 message=None,
                 duration=1234,
                 decoration=False,
                 geometry=None,
                 cnfs=None):
        """
        PARAMETERS:

        - master: widget parent. Example: an instance of tk.Frame

        - title: title of dialog box

        - header: the text to show as header

        - message: the text to show as message

        - duration: int, in milliseconds.
            You can set None to duration to cancel the self-destroying timer

        - decoration: True or False to allow Window decoration

        - geometry: str, as the dialog box is a toplevel (BODY),
         you can edit its geometry. Example: "500x300"

        - options: dictionary of widgets options
            The widgets keys are: BODY, LABEL_HEADER, LABEL_MESSAGE.

            Example: Assume that you want to set the LABEL_MESSAGE's background to black
            and the BODY's background to red:
                options = { BODY: {"background": "red"},
                            LABEL_MESSAGE: {"background": "black"} }
        """
        self.__cnfs = merge_cnfs(None, cnfs, components=("body",
                        LABEL_HEADER, LABEL_MESSAGE))
        super().__init__(master=master,
                         class_="Toast",
                         cnf=self.__cnfs["body"],
                         on_build=self.__on_build,
                         on_display=self.__on_display,
                         on_destroy=self.__on_destroy,
                         toplevel_geometry=self.__toplevel_geometry)
        self.__title = title
        self.__header = header
        self.__message = message
        self.__duration = duration
        self.__decoration = decoration
        self.__geometry = geometry
        self.__cancel_id = None
        self.__components = {}
        self.__view = self.build()

    # ======================================
    #            PROPERTIES
    # ======================================

    @property
    def header(self):
        return self.__header

    @property
    def message(self):
        return self.__message

    @property
    def duration(self):
        return self.__duration

    @property
    def decoration(self):
        return self.__decoration

    @property
    def components(self):
        """
        Get the components (widgets instances) used to build this Toast.

        This property returns a dict. The keys are:
            BODY, LABEL_HEADER, LABEL_MESSAGE,
        """
        return self.__components

    # ======================================
    #            LIFECYCLE
    # ======================================
    def __on_build(self):
        if not self.__decoration:
            self.overrideredirect(1)
        if self.__geometry:
            self.geometry(self.__geometry)
        if self.__title:
            self.title(self.__title)
        self.bind("<Button-1>", self.__on_click, "+")
        if self.__header:
            label_header = tk.Label(self,
                                    name=LABEL_HEADER,
                                    text=self.__header,
                                    anchor="w",
                                    justify=tk.LEFT,
                                    cnf=self.__cnfs[LABEL_HEADER])
            self.__components[LABEL_HEADER] = label_header
            label_header.pack(fill=tk.X, padx=10, pady=10)
        if self.__message:
            label_message = tk.Label(self,
                                     name=LABEL_MESSAGE,
                                     text=self.__message,
                                     anchor="w",
                                     justify=tk.LEFT,
                                     cnf=self.__cnfs[LABEL_MESSAGE])
            self.__components[LABEL_MESSAGE] = label_message
            label_message.pack(fill=tk.X, padx=10, pady=10)

    def __on_display(self):
        if self.__duration is not None:
            self.__cancel_id = self.after(self.__duration, self.destroy)

    def __on_destroy(self):
        pass

    def __toplevel_geometry(self):
        tkmisc.center_window(self, within=self.master.winfo_toplevel())
        tkmisc.dialog_effect(self)

    def __on_click(self, event):
        self.destroy()


class _ToastTest(View):

    def __init__(self, root):
        super().__init__()
        self._root = root
        self._body = None

    def _on_build(self):
        self._body = tk.Frame(self._root)
        btn_launch = tk.Button(self._body, text="launch",
                               command=self._on_click_launch)
        btn_launch.pack()

    def _on_display(self):
        pass

    def _on_destroy(self):
        pass

    def _on_click_launch(self):
        Toast(self._body, title="Toast Title",
              header="Header", message="This is the message",
              duration=3000)


if __name__ == "__main__":
    root = tk.Tk()
    root.geometry("500x300+0+0")
    toast_test = _ToastTest(root)
    toast_test.build_pack()
    root.mainloop()



from pyrustic.version import __version__
try:
    import importlib.metadata as dist_info
except ImportError:
    import importlib_metadata as dist_info


def dist(name, target=None):  # TODO: implement target=None or path to blah blah...
    """
    DESCRIPTION:
        Use this function to get some info about a distribution package
    PARAM:
        name: the distribution name, example: "wheel"
    RETURN:
        A dict with these keys:
            name, author, author_email, description, home_page,
            maintainer, maintainer_email, version
        All values are strings.
    """
    metadata_cache = None
    try:
        metadata_cache = dist_info.metadata(name)
    except Exception:
        pass
    keys = (("author", "Author"),
            ("author_email", "Author-email"),
            ("description", "Summary"),
            ("home_page", "Home-page"),
            ("maintainer", "Maintainer"),
            ("maintainer_email", "Maintainer-email"),
            ("version", "Version"))
    data = None
    if metadata_cache:
        data = {"name": name}
        for item in keys:
            if item[1] in metadata_cache:
                data[item[0]] = metadata_cache[item[1]]
    return data
